from __future__ import annotations

import logging
from collections.abc import Collection, Mapping
from typing import Any, cast

try:
    from typing import Self
except ImportError:
    try:
        from typing_extensions import Self
    except ImportError:
        ...
import gymnasium as gym
import numpy as np
import yaml
from navground import core, sim

from ..config import GroupConfig, StateConfig, ActionConfig, ObservationConfig
from ..policies.info_predictor import InfoPolicy
from ..types import Action, Array, Bounds, Observation
from .clock import SyncClock
from .group import Agent, GymAgent, create_agents_in_groups

StepReturn = tuple[dict[int, Observation], dict[int, float], dict[int, bool],
                   dict[int, bool], dict[int, dict[str, Action]]]
ResetReturn = tuple[dict[int, Observation], dict[int, dict[str, Action]]]


def make_scenario(
        value: sim.Scenario | str | dict[str, Any] | None
) -> sim.Scenario | None:
    if isinstance(value, dict):
        value = yaml.dump(value)
    if isinstance(value, str):
        return sim.load_scenario(value)
    return value


class NavgroundBaseEnv:
    """
    This class describes an environment that uses
    a :py:class:`navground.sim.Scenario` to
    generate and then simulate a :py:class:`navground.sim.World`.

    :param scenario: The scenario to initialize all simulated worlds.
                     If a :py:class:`str`, it will be interpreted as the YAML
                     representation of a scenario.
                     If a :py:class:`dict`, it will be dumped to YAML and
                     then treated as a :py:class:`str`.

    :param groups: The configuration of groups of agents controlled by the environment.
                   All other agents are controlled solely by navground.

    :param max_number_of_agents: The maximal number of agents that we will expose.
                                 It needs to be specified only for scenarios
                                 that generate world with a variable number of agents.

    :param time_step: The simulation time step applied at every :py:meth:`_step`.

    :param max_duration: If positive, it will signal a truncation after this simulated time.

    :param terminate_if_idle: Whether to terminate when an agent is idle

    :param truncate_outside_bounds: Whether to truncate when an agent exit the bounds

    :param bounds: The area to render and a fence for truncating processes when agents exit it.

    :param render_mode: The render mode.
                        If `"human"`, it renders a simulation in real time via
                        websockets (see :py:class:`navground.sim.ui.WebUI`).
                        If `"rgb_array"`, it uses
                        :py:func:`navground.sim.ui.render.image_for_world`
                        to render the world on demand.

    :param render_kwargs: Arguments passed to :py:func:`navground.sim.ui.render.image_for_world`

    :param realtime_factor: a realtime factor for `render_mode="human"`: larger values
                            speed up the simulation.

    :param stuck_timeout: The time to wait before considering an agent as stuck and terminate it.

    :param wait: Whether to signal termination/truncation only when
                 all agents have terminated/truncated.

    :param truncate_fast: Whether to signal truncation for all agents
                          as soon as one agent truncates.
    :param state: An optional global state configuration

    :param include_action: Whether to include field "navground_action" in the info

    :param include_success: Whether to include field "is_success" in the info

    :param init_success: The default value of success (valid until a termination condition is met)

    :param intermediate_success: Whether to include "is_success" in the info at intermediate
                                 steps (vs only at termination).

    """

    metadata = {"render_modes": ["human", "rgb_array"], 'render_fps': 30}
    MAX_SEED = 2**31 - 1
    action_key = 'navground_action'

    def __init__(self,
                 scenario: sim.Scenario | str | dict[str, Any] | None = None,
                 groups: Collection[GroupConfig] = tuple(),
                 max_number_of_agents: int | None = None,
                 time_step: float = 0.1,
                 max_duration: float = -1.0,
                 terminate_if_idle: bool = True,
                 bounds: Bounds | None = None,
                 truncate_outside_bounds: bool = False,
                 render_mode: str | None = None,
                 render_kwargs: Mapping[str, Any] = {},
                 realtime_factor: float = 1.0,
                 stuck_timeout: float = -1,
                 wait: bool = False,
                 truncate_fast: bool = False,
                 include_action: bool = True,
                 include_success: bool = True,
                 init_success: bool | None = None,
                 intermediate_success: bool = False,
                 state: StateConfig | None = None) -> None:

        assert render_mode is None or render_mode in self.metadata[
            "render_modes"]  # type: ignore
        self.groups_config = groups
        self.time_step = time_step
        self.max_duration = max_duration
        self.terminate_if_idle = terminate_if_idle
        self.bounds = bounds
        self.truncate_outside_bounds = truncate_outside_bounds
        self.render_mode = render_mode
        self.render_kwargs = render_kwargs
        self.realtime_factor = realtime_factor
        self.stuck_timeout = stuck_timeout
        self.wait = wait
        self.truncate_fast = truncate_fast
        self.include_action = include_action
        self.include_success = include_success
        self.intermediate_success = intermediate_success
        self.init_success = init_success
        self._scenario: sim.Scenario | None = make_scenario(scenario)
        self.max_number_of_agents = max_number_of_agents
        self.state_config = state
        self._truncatated = False
        self._terminated = False
        self._spec: dict[str, Any]
        self._init_spec(scenario)
        self._init()

    def action_config(self, index: int) -> ActionConfig | None:
        """
        Gets the action configuration for a (possible) agent

        :param      index:  The agent index

        :returns:   The action configuration, or None if undefined.
        """
        if index in self._possible_agents:
            g = self._possible_agents[index].gym
            return cast('ActionConfig', g.action_config) if g else None
        return None

    def get_cmd_from_action(self, index: int, action: Action,
                            time_step: float) -> core.Twist2 | None:
        """
        Convert action to navground command for a (possible) agent

        :param      index:  The agent index
        :param      action:     The action
        :param      time_step:  The time step

        :returns:   A control command or None if no agent is configured
                    at the given index.
        """
        if index in self._possible_agents:
            g = self._possible_agents[index].gym
            return g.get_cmd_from_action(action, time_step) if g else None
        return None

    def observation_config(self, index: int) -> ObservationConfig | None:
        """
        Gets the observation configuration for a (possible) agent

        :param      index:  The agent index

        :returns:   The observation configuration, or None if undefined.
        """
        if index in self._possible_agents:
            g = self._possible_agents[index].gym
            return cast('ObservationConfig',
                        g.observation_config) if g else None
        return None

    def _init_spec(
            self,
            scenario: sim.Scenario | str | dict[str, Any] | None) -> None:
        self._spec = {
            'max_number_of_agents': self.max_number_of_agents,
            'groups': self.groups_config,
            'scenario': scenario,
            'time_step': self.time_step,
            'terminate_if_idle': self.terminate_if_idle,
            'bounds': self.bounds,
            'max_duration': self.max_duration,
            'truncate_outside_bounds': self.truncate_outside_bounds,
            'render_mode': self.render_mode,
            'render_kwargs': self.render_kwargs,
            'realtime_factor': self.realtime_factor,
            'stuck_timeout': self.stuck_timeout,
            'wait': self.wait,
            'truncate_fast': self.truncate_fast,
            'include_action': self.include_action,
            'include_success': self.include_success,
            'init_success': self.init_success,
            'intermediate_success': self.intermediate_success,
            'state': self.state_config
        }

    def _init(self) -> None:
        self._world: sim.World | None = None
        self._agents: dict[int, Agent] = {}
        self._termination: dict[int, bool] = {}
        self._truncation: dict[int, bool] = {}
        self._success: dict[int, bool] = {}
        self._loop: asyncio.AbstractEventLoop | None = None
        self._state_space: gym.spaces.Box | None = None
        if self._scenario is None:
            self._possible_agents: dict[int, Agent] = {}
        else:
            world = self._scenario.make_world()
            if self.state_config:
                self._state_space = self.state_config.get_space(world)
            self._possible_agents = {
                i: agent
                for i, agent in create_agents_in_groups(
                    world, self.groups_config,
                    self.max_number_of_agents).items() if agent.gym is not None
            }
        self._observation_space = {
            i: cast("GymAgent", agent.gym).observation_space
            for i, agent in self._possible_agents.items()
        }
        self._action_space = {
            i: cast("GymAgent", agent.gym).action_space
            for i, agent in self._possible_agents.items()
        }
        if self.render_mode == "human" and not self._loop:
            import asyncio

            from navground.sim.ui import WebUI

            self._loop = asyncio.get_event_loop()
            self._rt_clock: SyncClock | None = SyncClock(
                self.time_step, factor=self.realtime_factor)
            self._ui: WebUI | None = WebUI(port=8002)
            self._loop.run_until_complete(self._ui.prepare())
        else:
            self._loop = None
            self._ui = None
            self._rt_clock = None
        # self._observation_item_space = {
        #     i: cast(GymAgent, agent.gym).observation_item_space
        #     for i, agent in self._possible_agents.items()
        # }

    # def get_observation_item_space(self, index: int) -> gym.spaces.Dict | None:
    #     return self._observation_item_space.get(index)

    def update_agents(self) -> None:
        if not self._world:
            return
        self._agents.clear()
        for group in self.groups_config:
            indices = group.indices.as_set(len(self._world.agents))
            for i in indices:
                ng_agent = self._world.agents[i]
                self._agents[i] = self._possible_agents[i]
                self._agents[i].navground = ng_agent
                gym_agent = self._agents[i].gym
                if gym_agent:
                    gym_agent.init(ng_agent.behavior, self._agents[i].state)
        self._termination = {i: False for i in self._agents}
        self._truncation = {i: False for i in self._agents}
        self._success = {}

    @property
    def init_args(self) -> dict[str, Any]:
        """
        Returns the arguments used to initialize the environment

        :returns:   The initialization arguments.
        """
        return self._spec

    @property
    def asdict(self) -> dict[str, Any]:
        """
        A JSON-able representation of the instance

        :returns:  A JSON-able dict
        """
        rs = {
            'groups': [g.asdict for g in self.groups_config],
            'time_step': self.time_step,
            'max_duration': self.max_duration,
            'terminate_if_idle': self.terminate_if_idle,
            'truncate_outside_bounds': self.truncate_outside_bounds,
            'render_mode': self.render_mode,
            'render_kwargs': self.render_kwargs,
            'realtime_factor': self.realtime_factor,
            'stuck_timeout': self.stuck_timeout,
            'wait': self.wait,
            'truncate_fast': self.truncate_fast,
            'include_action': self.include_action,
            'include_success': self.include_success,
            'init_success': self.init_success,
            'intermediate_success': self.intermediate_success
        }
        if self.state_config:
            rs['state'] = self.state_config.asdict
        if self._scenario:
            rs['scenario'] = yaml.safe_load(sim.dump(self._scenario))
        if self.bounds:
            rs['bounds'] = [p.tolist() for p in self.bounds]
        rs['type'] = type(self).__name__
        return rs

    @classmethod
    def from_dict(cls, value: Mapping[str, Any]) -> Self:
        """
        Load the class from the JSON representation

        :param value:  A JSON-able dict
        :returns: An instance of the class
        """
        t = value.get('type', '')
        for subcls in cls.__subclasses__():
            if subcls.__name__ == t:
                cls = subcls
                break
        env = cls()
        env.groups_config = [
            GroupConfig.from_dict(v) for v in value.get('groups', [])
        ]
        env.time_step = value.get('time_step', 0.1)
        env.max_duration = value.get('max_duration', -1)
        env.terminate_if_idle = value.get('terminate_if_idle', True)
        env.truncate_outside_bounds = value.get('truncate_outside_bounds',
                                                False)
        env.render_mode = value.get('render_mode')
        env.render_kwargs = value.get('render_kwargs', {})
        env.realtime_factor = value.get('realtime_factor', 1)
        env.stuck_timeout = value.get('stuck_timeout', -1)
        env.wait = value.get('wait', False)
        env.truncate_fast = value.get('truncate_fast', False)
        env.include_action = value.get('include_action', None)
        env.init_success = value.get('init_success', None)
        env.include_success = value.get('include_success', True)
        env.intermediate_success = value.get('intermediate_success', True)
        if 'state' in value:
            env.state_config = StateConfig.from_dict(value['state'])
        else:
            env.state_config = None
        if 'scenario' in value:
            env._scenario = sim.load_scenario(yaml.safe_dump(
                value['scenario']))
        else:
            env._scenario = None
        env.bounds = None
        if 'bounds' in value:
            ps = value['bounds']
            assert len(ps) == 2
            p1, p2 = ps
            env.bounds = (np.asarray(p1), np.asarray(p2))
        env._init()
        env._init_spec(value.get('scenario'))
        return env

    @property
    def scenario(self) -> sim.Scenario | None:
        """
        The navground scenario, if set.
        """
        return self._scenario

    def _reset(self,
               seed: int | None = None,
               options: dict[str, Any] | None = None) -> ResetReturn:
        world = self._world
        if world:
            world._close()
        self._world = sim.World()
        if seed is None and world:
            self._world.copy_random_generator(world)
        if seed is not None:
            seed = seed & self.MAX_SEED
        if not self._scenario:
            return {}, {}
        self._scenario.init_world(self._world, seed=seed)
        self._scenario.apply_inits(self._world)
        self._world._prepare()
        # Update dry does change last_cmd. Let's cache it to restore it afterwards
        last_cmds = [a.last_cmd for a in self._world.agents]
        self._world.update_dry(self.time_step, advance_time=False)
        self.update_agents()
        self._truncated = False
        self._terminated = False
        obs = self.get_observations()
        if self.render_mode == "human" and self._ui and self._loop:
            self._loop.run_until_complete(
                self._ui.init(self._world, bounds=self.bounds))
        infos = self.get_infos()
        # ... restore last_cmd
        for a, cmd in zip(self._world.agents, last_cmds, strict=True):
            a.last_cmd = cmd
        return obs, infos

    def _step(self, actions: dict[int, Action]) -> StepReturn:
        assert self._world is not None
        try:
            for index, action in actions.items():
                if self.agent_is_active(index):
                    agent = self._agents[index]
                    if agent.navground and agent.gym:
                        cmd = agent.gym.get_cmd_from_action(
                            action, self.time_step)
                        if agent.navground.behavior:
                            agent.navground.last_cmd = agent.navground.behavior.feasible_twist(
                                cmd)
        except Exception as e:
            logging.warning(e)
            pass
        self._world.actuate(self.time_step)
        self._world.update_dry(self.time_step, advance_time=False)
        if self.render_mode == "human" and self._ui and self._rt_clock and self._loop:
            self._loop.run_until_complete(self._ui.update(self._world))
            self._rt_clock.tick()
        rewards = self.get_rewards()
        self.update_termination()
        self.update_truncation()
        return (self.get_observations(), rewards,
                self.get_termination(), self.get_truncation(),
                self.get_infos())

    def get_rewards(self) -> dict[int, float]:
        if self._world:
            return {
                index:
                agent.reward(agent.navground, self._world, self.time_step)
                if self.agent_is_active(index) else 0
                for index, agent in self._agents.items()
                if agent.navground and agent.reward
            }
        return {}

    def render(self) -> Any:
        if self.render_mode == "rgb_array" and self._world:
            from navground.sim.ui.render import image_for_world

            return image_for_world(self._world,
                                   bounds=self.bounds,
                                   **self.render_kwargs)

    def close(self) -> None:
        if self._ui and self._loop:
            self._loop.run_until_complete(self._ui.stop())
            self._ui = None

    def get_observations(self) -> dict[int, Observation]:
        if self._world:
            for agent in self._agents.values():
                agent.update_state(self._world)
        return {
            index: agent.gym.update_observation()
            for index, agent in self._agents.items() if agent.gym
        }

    # Not truncated and not terminated
    def agent_is_active(self, index: int) -> bool:
        return not self.has_terminated(index) and not self.has_truncated(index)

    def update_success(self) -> None:
        if not self._world:
            return
        for index, agent in self._agents.items():
            if not self._termination[index]:
                if agent.navground:
                    if agent.success_condition:
                        if agent.success_condition(agent.navground,
                                                   self._world):
                            if index not in self._success:
                                self._success[index] = True
                            if agent.terminate_on_success:
                                self.terminate(index)
                                continue
                    if agent.failure_condition:
                        if agent.failure_condition(agent.navground,
                                                   self._world):
                            if index not in self._success:
                                self._success[index] = False
                            if agent.terminate_on_failure:
                                self.terminate(index)
                                continue

    def terminate(self, index: int):
        self._termination[index] = True
        if all(self._termination.values()):
            self._terminated = True

    def has_terminated(self, index: int) -> bool:
        if self.wait:
            return self._terminated
        return self._termination.get(index, False)

    def get_termination(self) -> dict[int, bool]:
        if self.wait and not self._terminated:
            return {i: False for i in self._agents}
        return self._termination

    def truncate(self, index: int):
        self._truncation[index] = True
        self._truncated = True

    def get_truncation(self) -> dict[int, bool]:
        # if self.wait and not all(self._truncation.values()):
        #     return {i: False for i in self._truncation}
        if self.truncate_fast and self._truncated:
            return {i: True for i in self._agents}
        return self._truncation

    def has_truncated(self, index: int) -> bool:
        if self.truncate_fast:
            return self._truncated
        return self._truncation.get(index, False)

    def update_termination(self) -> None:
        if not self._world:
            return
        self.update_success()
        if self._world.should_terminate():
            self._terminated = True
            return
        time = self._world.time
        for index, agent in self._agents.items():
            if not self._termination[index]:
                if agent.navground:
                    if self.terminate_if_idle and agent.navground.idle:
                        self.terminate(index)
                        continue
                    if self.stuck_timeout > 0 and agent.navground.has_been_stuck_since(
                            time - self.stuck_timeout):
                        self.terminate(index)
                        continue

    def update_truncation(self):
        if self._world and self.max_duration > 0 and self.max_duration <= self._world.time:
            self._truncated = True
            self._truncation = {i: True for i in self._agents}
            return
        out = self.should_check_if_outside_bounds
        for index, agent in self._agents.items():
            if self._termination[index] or self._truncation[index]:
                continue
            if agent.navground:
                if out and self.is_outside_bounds(agent.navground.position):
                    self.truncate(index)

    @property
    def should_check_if_outside_bounds(self) -> bool:
        return self.truncate_outside_bounds and self.bounds is not None

    def is_outside_bounds(self, position: core.Vector2) -> bool:
        if self.bounds is not None:
            return any(position < self.bounds[0]) or any(
                position > self.bounds[1])
        return True

    def get_agent_infos(self, index: int, agent: Agent) -> dict[str, Action]:
        info = {}
        if self.include_action and agent.gym:
            if self.agent_is_active(index):
                info[self.action_key] = agent.gym.get_action(self.time_step)
            else:
                space = self._action_space[index]
                if isinstance(space, gym.spaces.Box):
                    info[self.action_key] = np.zeros_like(space.low)
        if self.include_success:
            if not self.agent_is_active(index) or self.intermediate_success:
                success = self._success.get(index, self.init_success)
                if success is not None:
                    info['is_success'] = success  # type: ignore
        return info

    def get_infos(self) -> dict[int, dict[str, Action]]:
        return {
            index: self.get_agent_infos(index, agent)
            for index, agent in self._agents.items() if agent.gym
        }

    def get_policy(self, index: int) -> InfoPolicy:
        """
        A policy that returns the action computed by the navground agent.

        :returns:   The policy.
        """
        return InfoPolicy(action_space=self._action_space[index],
                          observation_space=self._observation_space[index],
                          key=self.action_key)

    def get_state(self) -> Array | None:
        if self._world and self.state_config:
            return self.state_config.get_state(self._world)
        return None

    @property
    def state_space(self) -> gym.spaces.Box | None:
        return self._state_space

    @property
    def has_state(self) -> bool:
        return self.state_config is not None



from .discrete_modulation import DiscreteModulationActionConfig
from .policy_behavior_modulation import PolicyModulation, add_modulation
from .selection_behavior import SelectionBehavior

__all__ = [
    "DiscreteModulationActionConfig", "PolicyModulation", "add_modulation",
    "SelectionBehavior"
]



from __future__ import annotations

from pathlib import Path
from typing import Any

import pandas as pd

from ...types import PathLike
from ..plot import LogField, plot_logs


def load_logs(path: PathLike, key: str) -> pd.DataFrame:
    """
    Loads logs from a csv.

    :param      path:  The directory with csv logs
    :param      key:   The prefix of the fields to load

    :returns:   A dataframe with the logs
    """
    progress = pd.read_csv(Path(path) / 'progress.csv')
    cols = [k for k in progress.columns if key in k]
    keys = ['time/total_timesteps'] + cols
    fs = ~progress[cols[0]].isna()
    f_progress = progress[keys].loc[fs]
    return f_progress.fillna(0)


def load_eval_logs(path: PathLike) -> pd.DataFrame:
    """
    Loads the evaluation logs from a csv.

    :param      path:  The directory with csv logs

    :returns:   A dataframe with the logs
    """
    return load_logs(path, key='eval')


def load_rollout_logs(path: PathLike) -> pd.DataFrame:
    """
    Loads the rollout logs from a csv.

    :param      path:  The directory with csv logs

    :returns:   A dataframe with the logs
    """
    return load_logs(path, key='rollout')


def plot_eval_logs(path: PathLike,
                   two_axis: bool = False,
                   reward: bool = True,
                   reward_color: str = 'k',
                   reward_linestyle: str = '--',
                   success: bool = False,
                   length: bool = False,
                   reward_low: float | None = None,
                   reward_high: float = 0,
                   length_low: float = 0,
                   length_high: float | None = None,
                   **kwargs: Any) -> None:
    """
    Plots reward, success and/or length from the eval logs

    :param      path:         The directory with csv logs
    :param      reward:       Whether to plot the reward
    :param      reward_color: The reward color
    :param      reward_linestyle: The reward linestyle
    :param      success:      Whether to plot the success
    :param      length:       Whether to plot the length
    :param      reward_low:   An optional lower bound to scale the reward
    :param      reward_high:  An optional upper bound to scale the reward
    :param      lenght_low:   An optional lower bound to scale the reward
    :param      lenght_high:  An optional upper bound to scale the reward
    :param      two_axis:     Whether to use two axis (only if there are two fields)
    """
    fields: list[LogField] = []
    if reward:
        fields.append(
            LogField(label="mean reward",
                     key="eval/mean_reward",
                     linestyle=reward_linestyle,
                     color=reward_color,
                     low=reward_low,
                     high=reward_high))
    if success:
        fields.append(
            LogField(label="success rate",
                     key="eval/success_rate",
                     linestyle='-',
                     color='g'))
    if length:
        fields.append(
            LogField(label="mean length",
                     key="eval/mean_ep_length",
                     linestyle=':',
                     color='b',
                     low=length_low,
                     high=length_high))
    plot_logs(load_logs(path, key='eval'),
              title="eval",
              two_axis=two_axis,
              fields=fields,
              key='time/total_timesteps',
              **kwargs)


def plot_rollout_logs(path: PathLike,
                      two_axis: bool = False,
                      reward: bool = True,
                      success: bool = False,
                      length: bool = False,
                      reward_low: float | None = None,
                      reward_high: float = 0,
                      length_low: float = 0,
                      length_high: float | None = None,
                      **kwargs: Any) -> None:
    """
    Plots reward, success and/or length from the rollout logs

    :param      path:         The directory with csv logs
    :param      reward:       Whether to plot the reward
    :param      success:      Whether to plot the success
    :param      length:       Whether to plot the length
    :param      reward_low:   An optional lower bound to scale the reward
    :param      reward_high:  An optional upper bound to scale the reward
    :param      lenght_low:   An optional lower bound to scale the reward
    :param      lenght_high:  An optional upper bound to scale the reward
    :param      two_axis:     Whether to use two axis (only if there are two fields)
    """
    fields: list[LogField] = []
    if reward:
        fields.append(
            LogField(label="mean reward",
                     key="rollout/ep_rew_mean",
                     linestyle='--',
                     color='k',
                     low=reward_low,
                     high=reward_high))
    if success:
        fields.append(
            LogField(label="success rate",
                     key="rollout/success_rate",
                     linestyle='-',
                     color='g'))
    if length:
        fields.append(
            LogField(label="mean length",
                     key="rollout/ep_len_mean",
                     linestyle=':',
                     color='b',
                     low=length_low,
                     high=length_high))
    plot_logs(load_logs(path, key='rollout'),
              title="rollout",
              two_axis=two_axis,
              fields=fields,
              key='time/total_timesteps',
              **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from .utils import get_number_of_batches

if TYPE_CHECKING:
    import gymnasium as gym

    from ..types import Action, EpisodeStart, Observation, State


class NullPredictor:
    """
    This class describes a dummy predictor that always returns zeros and
    conforms to :py:class:`navground.learning.types.PolicyPredictor`

    :param action_space:       The action space
    :param observation_space:  An optional observation space
    """

    def __init__(self, action_space: gym.spaces.Box,
                 observation_space: gym.Space[Any]) -> None:
        self.observation_space = observation_space
        self.action_space = action_space

    def get_actions(
        self,
        observation: Observation,
    ) -> Action:
        number = get_number_of_batches(observation, self.observation_space)
        if number == 0:
            return np.zeros(shape=self.action_space.shape,
                            dtype=self.action_space.dtype)
        return np.zeros(
            shape=(number, *self.action_space.shape),
            dtype=self.action_space.dtype)

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False) -> tuple[Action, State | None]:
        return self.get_actions(observation), None

    def __call__(
        self,
        observation: Observation,
        state: State | None = None,
        episode_start: EpisodeStart | None = None
    ) -> tuple[Action, State | None]:
        return self.predict(observation, state, episode_start, True)



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import Any, TypeVar, cast, TYPE_CHECKING

import gymnasium as gym
import numpy as np
from navground import core

from ..env.env import BaseEnv
from ..types import Action, Array, Observation

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv

T = TypeVar('T', bound=gym.Space[Any])


def stack_infos(values: dict[int, dict[str, Action]],
                all_of: set[str] = {'is_success'}) -> dict[str, Action]:
    if not values:
        return {}
    rs: dict[str, Any] = {}
    keys: set[str] = set([v for value in values.values() for v in value])
    keys -= all_of
    for k in all_of:
        ms = set(value[k] for value in values.values() if k in value)
        if any(m is False for m in ms):
            rs[k] = False
        elif ms and all(m is True for m in ms):
            rs[k] = True
    for k in keys:
        ns = [value[k] for value in values.values() if k in value]
        rs[k] = np.stack(ns)
    return rs


def stack_observations(
        values: dict[int, Observation],
        dtype: type[np.floating[Any]] = core.FloatType) -> Observation:
    if not values:
        return np.array([], dtype=dtype)
    value = next(iter(values.values()))
    if isinstance(value, Mapping):
        keys = value.keys()
        return {k: np.stack([v[k] for v in values.values()]) for k in keys}
    return np.stack(list(cast('dict[int, Array]', values).values()))


def stack_box_spaces(spaces: list[gym.spaces.Box]):
    low = np.stack([space.low for space in spaces])
    high = np.stack([space.high for space in spaces])
    return gym.spaces.Box(low, high)


def stack_observation_spaces(values: dict[int, T]) -> T:
    assert (values)
    value = next(iter(values.values()))
    if isinstance(value, Mapping):
        keys = value.keys()
        spaces = cast('Iterable[Mapping[str, gym.spaces.Box]]',
                      values.values())
        return gym.spaces.Dict({
            k:
            stack_box_spaces([space[k] for space in spaces])
            for k in keys
        })
    return stack_box_spaces(
        list(cast('dict[int, gym.spaces.Box]', values).values()))


def unstack_actions(values: Array, agents: list[int]) -> dict[int, Array]:
    return dict(zip(agents, values, strict=True))


def get_state_space(env: ParallelEnv) -> gym.spaces.Box | None:
    """
    Gets the state space, first by trying to read
    ``env.state_space`` directly and then from ``env.state``.

    :param env:   The environment

    :returns:   The state space or None if not defined.
    """

    try:
        return env.state_space  # type: ignore
    except AttributeError:
        pass
    env.reset()
    state = env.state()
    return gym.spaces.Box(low=-np.inf,
                          high=np.inf,
                          shape=state.shape,
                          dtype=state.dtype)


class JointEnv(gym.Env):
    """
    Wraps a multi-agent parallel environment as a single
    agent environment, stacking observation but aggregating
    rewards, terminations and truncations.

    Requires that *homogeneous* observations spaces (if ``state=False``) and
    action spaces.

    :param env:   The parallel environment
    :param state: Whether to return the global state as observations
                  (vs the stacked observation).
    """

    def __init__(self, env: ParallelEnv, state: bool = False) -> None:
        self.env = env
        self.use_state = False
        if state:
            self.use_state = True
            state_space = get_state_space(env)
            if state_space:
                self.observation_space = state_space
            else:
                raise ValueError("Cannot get state space")
        else:
            self.use_state = False
            observation_spaces = {
                i: env.observation_space(i)
                for i in env.possible_agents
            }
            self.observation_space = stack_observation_spaces(
                observation_spaces)
        action_spaces = cast(
            'list[gym.spaces.Box]',
            [env.action_space(i) for i in env.possible_agents])
        self.action_space = stack_box_spaces(action_spaces)

    @property
    def unwrapped(self) -> BaseEnv:
        return self.env.unwrapped  # type: ignore

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None
    ) -> tuple[Observation, dict[str, Any]]:
        """
        Conforms to :py:meth:`gymnasium.Env.reset`.

        Resets the parallel environment and return stacked
        observation (or the global state, if :py:attr:`use_state` is true)
        and infos.
        """
        obs, infos = self.env.reset(seed=seed, options=options)
        if self.use_state:
            state: Observation | None = self.env.state()
            if not state:
                raise RuntimeError("Environment has no state")
            vobs = state
        else:
            vobs = stack_observations(obs)
        return vobs, stack_infos(infos)

    def step(
        self, action: Action
    ) -> tuple[Observation, float, bool, bool, dict[str, Array]]:
        """
        Conforms to :py:meth:`gymnasium.Env.step`.

        Converts the action to an agent-indexed dictionary
        ``[act_0, act_1, ...]`` -> ``{0: act_0, 1: act_1, ...}``
        and forwards them to :py:meth:`pettingzoo.utils.env.ParallelEnv.step`.

        Returns stacked observations (or the global state, if :py:attr:`use_state` is true)
        and infos, and aggregated reward (sum), termination (all) , truncation (any).
        """
        obs, reward, terminated, truncated, infos = self.env.step(
            unstack_actions(action, self.env.agents))
        if self.use_state:
            state: Observation | None = self.env.state()
            if not state:
                raise RuntimeError("Environment has no state")
            vobs = state
        else:
            vobs = stack_observations(obs)
        return (vobs, sum(x for _, x in reward.items()),
                all(x for _, x in terminated.items()),
                any(x for _, x in truncated.items()), stack_infos(infos))



from __future__ import annotations

import numpy as np
from navground import core, sim


class ForwardScenario(sim.Scenario, name='Forward'):
    """
    Periodic worlds with agents traveling in (constant) random
    target directions between random obstacles:

    :param      width:                    The width of the (periodic) cell
    :param      length:                   The length of the (periodic) cell
    :param      min_number_of_obstacles:  The minimum number of obstacles
    :param      max_number_of_obstacles:  The maximum number of obstacles
    :param      min_obstacle_radius:      The minimal obstacles radius
    :param      max_obstacle_radius:      The maximal obstacles radius
    :param      margin:                   The margin between agents at init
    :param      periodic_x:               Whether the world is periodic along the x-axis
    :param      periodic_y:               Whether the world is periodic along the y-axis

    """

    def __init__(
        self,
        width: float = 2.0,
        length: float = 6.0,
        min_number_of_obstacles: int = 0,
        max_number_of_obstacles: int = 0,
        min_obstacle_radius: float = 0.2,
        max_obstacle_radius: float = 0.2,
        margin: float = 0.0,
        periodic_x: bool = True,
        periodic_y: bool = True,
    ) -> None:
        super().__init__()
        self._width = width
        self._length = length
        self._min_number_of_obstacles = min_number_of_obstacles
        self._max_number_of_obstacles = max_number_of_obstacles
        self._min_obstacle_radius = min_obstacle_radius
        self._max_obstacle_radius = max_obstacle_radius
        self._margin = margin
        self._periodic_x = periodic_x
        self._periodic_y = periodic_y

    def init_world(self, world: sim.World, seed: int | None = None) -> None:
        """
        Initializes the world.

        :param      world:  The world
        :param      seed:   The seed

        """

        sim.Scenario.init_world(self, world, seed)
        rng = world.random_generator
        # rng = np.random.default_rng(seed)
        world.bounding_box = sim.BoundingBox(0, self.length, -self.width * 0.5,
                                             self.width * 0.5)
        if self.periodic_x:
            world.set_lattice(0, (0, self.length))
        if self.periodic_y:
            world.set_lattice(1, (-self.width * 0.5, self.width))

        for i, agent in enumerate(world.agents):
            if i == 0:
                e = (1, 0)
                p = core.Pose2((0, 0), rng.uniform(high=2 * np.pi))
            else:
                alpha = rng.uniform(high=2 * np.pi)
                e = (np.cos(alpha), np.sin(alpha))
                p = core.Pose2((rng.uniform(low=0, high=self.length),
                                rng.uniform(low=-self.width * 0.5,
                                            high=self.width * 0.5)), alpha)
            agent.controller.follow_direction(e)
            agent.pose = p

        world.space_agents_apart(self.margin, with_safety_margin=True)
        margin = self.margin
        if world.agents:
            margin += 2 * max(
                agent.behavior.safety_margin if agent.behavior else 0 +
                agent.radius for agent in world.agents)

        if self._min_number_of_obstacles >= self._max_number_of_obstacles:
            number_of_obstacles = self._max_number_of_obstacles
        else:
            number_of_obstacles = rng.integers(
                low=self._min_number_of_obstacles,
                high=self._max_number_of_obstacles)
        if number_of_obstacles > 0:
            world.add_random_obstacles(number_of_obstacles,
                                       min_radius=self.min_obstacle_radius,
                                       max_radius=self.max_obstacle_radius,
                                       margin=margin,
                                       max_tries=10_000)

    @property
    @sim.register(6.0, "length of the area")
    def length(self) -> float:
        return self._length

    @length.setter
    def length(self, value: float) -> None:
        self._length = max(0, value)

    @property
    @sim.register(2.0, "width of the area")
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, value: float) -> None:
        self._width = max(0, value)

    @property
    @sim.register(0, "min number of obstacles")
    def min_number_of_obstacles(self) -> int:
        return self._min_number_of_obstacles

    @min_number_of_obstacles.setter
    def min_number_of_obstacles(self, value: int) -> None:
        self._min_number_of_obstacles = max(0, value)

    @property
    @sim.register(0, "max number of obstacles")
    def max_number_of_obstacles(self) -> int:
        return self._max_number_of_obstacles

    @max_number_of_obstacles.setter
    def max_number_of_obstacles(self, value: int) -> None:
        self._max_number_of_obstacles = max(0, value)

    @property
    @sim.register(0.2, "min obstacle radius")
    def min_obstacle_radius(self) -> float:
        return self._min_obstacle_radius

    @min_obstacle_radius.setter
    def min_obstacle_radius(self, value: float) -> None:
        self._min_obstacle_radius = max(0, value)

    @property
    @sim.register(0.2, "max obstacle radius")
    def max_obstacle_radius(self) -> float:
        return self._max_obstacle_radius

    @max_obstacle_radius.setter
    def max_obstacle_radius(self, value: float) -> None:
        self._max_obstacle_radius = max(0, value)

    @property
    @sim.register(0.1, "margin")
    def margin(self) -> float:
        return self._margin

    @margin.setter
    def margin(self, value: float) -> None:
        self._margin = max(0, value)

    @property
    @sim.register(True, "periodic_x")
    def periodic_x(self) -> bool:
        return self._periodic_x

    @periodic_x.setter
    def periodic_x(self, value: bool) -> None:
        self._periodic_x = value

    @property
    @sim.register(True, "periodic_x")
    def periodic_y(self) -> bool:
        return self._periodic_y

    @periodic_y.setter
    def periodic_y(self, value: bool) -> None:
        self._periodic_y = value



from __future__ import annotations

import pathlib as pl
import warnings
from collections.abc import Iterator
from typing import TYPE_CHECKING, cast

try:
    from contextlib import chdir  # Python>=3.11
except ImportError:
    import contextlib
    import os

    @contextlib.contextmanager  # type: ignore[no-redef]
    def chdir(path: os.PathLike[str]) -> Iterator[None]:
        _old = os.getcwd()
        os.chdir(os.path.abspath(path))
        try:
            yield
        finally:
            os.chdir(_old)


import yaml
from navground import core, sim

from .behaviors.policy import PolicyBehavior
from .config import ControlActionConfig, DefaultObservationConfig
from .env import BaseEnv, NavgroundEnv
from .internal.base_env import NavgroundBaseEnv
from .internal.group import Agent
from .onnx.export import export
from .parallel_env import BaseParallelEnv, MultiAgentNavgroundEnv
from .types import PathLike

if TYPE_CHECKING:
    from stable_baselines3.common.base_class import BaseAlgorithm
    from stable_baselines3.common.policies import BasePolicy

    from .il import BaseILAlgorithm


def _get_agent(env: NavgroundBaseEnv,
               index: int | None = None) -> Agent | None:
    if index is None:
        if env._possible_agents:
            return next(iter(env._possible_agents.values()))
        return None
    return env._possible_agents.get(index, None)


def _get_policy_behavior(env: NavgroundBaseEnv,
                         index: int | None = None) -> PolicyBehavior | None:

    agent = _get_agent(env, index)
    if (agent and agent.navground and agent.gym and agent.navground.behavior
            and isinstance(agent.gym.action_config, ControlActionConfig)
            and isinstance(agent.gym.observation_config,
                           DefaultObservationConfig)):
        return PolicyBehavior.clone_behavior(
            behavior=agent.navground.behavior,
            action_config=agent.gym.action_config,
            observation_config=agent.gym.observation_config,
            policy=None)
    return None


def _get_sensors(env: NavgroundBaseEnv,
                 index: int | None = None) -> list[sim.Sensor]:
    agent = _get_agent(env, index)
    if agent:
        return agent.sensors
    return []


def save_env(
    env: BaseEnv | BaseParallelEnv,
    path: PathLike,
) -> None:
    """
    Export a NavgroundEnv to YAML

    :param      path:    The directory where to save the files
    :param      env:     The environment.
    :raises TypeError:   If ``env.unwrapped`` is not of a navground environment
    """
    if not isinstance(env.unwrapped, NavgroundBaseEnv):
        raise TypeError("Not a navground environment")
    with open(path, 'w') as f:
        f.write(yaml.safe_dump(env.unwrapped.asdict))


def load_env(path: PathLike) -> MultiAgentNavgroundEnv | NavgroundEnv:
    with open(path) as f:
        data = yaml.safe_load(f.read())
    return cast("MultiAgentNavgroundEnv | NavgroundEnv",
                NavgroundBaseEnv.from_dict(data))


def export_behavior(
    model: BaseAlgorithm | BaseILAlgorithm,
    path: PathLike,
) -> None:
    """
    Export a model using :py:func:`export_policy_as_behavior`

    :param      model:   The model
    :param      path:    The directory where to save the files
    """
    venv = model.env
    if venv:
        value = venv.get_attr("asdict", [0])[0]
        env: BaseEnv | BaseParallelEnv | None = cast(
            "BaseEnv | BaseParallelEnv", NavgroundBaseEnv.from_dict(value))
    else:
        env = None
    export_policy_as_behavior(path=path, policy=model.policy, env=env)


def export_policy_as_behavior(path: PathLike,
                              policy: BasePolicy | None = None,
                              env: BaseEnv | BaseParallelEnv | None = None,
                              index: int | None = None) -> None:
    """
    Export a policy (using :py:func:`navground.learning.onnx.export`)
    together with the YAML representation of the behavior and sensor to use it.

    The behavior and sensor refer to an agent at a given index in the environment.
    If they cannot be retrieved, the related files will not be saved.

    :param      path:    The directory where to save the files
    :param      policy:  The policy
    :param      env:     The environment.
    :param      index:   The index of the agent in the environment.
                         If not provided, selects one of the possible agents.
    """
    path = pl.Path(path)
    if path.is_file():
        warnings.warn(f"{path} points to a file", stacklevel=1)
        return
    if not path.exists():
        path.mkdir(parents=True)
    if policy:
        export(policy, path / "policy.onnx")
    if env and not isinstance(env, NavgroundBaseEnv):
        env = env.unwrapped
    if not (env and isinstance(env, NavgroundBaseEnv)):
        return
    behavior = _get_policy_behavior(env, index=index)
    if behavior:
        if policy:
            behavior.set_policy_path('policy.onnx', load_policy=False)
        with open(path / 'behavior.yaml', 'w') as f:
            f.write(behavior.dump())
    sensors = _get_sensors(env, index=index)
    sensors_as_dict = [yaml.safe_load(sensor.dump()) for sensor in sensors]
    if sensors:
        with open(path / 'sensor.yaml', 'w') as f:
            f.write(yaml.dump(sensors_as_dict))
    # TODO(Jerome): Maybe add a scenario


def load_behavior(
        path: PathLike
) -> tuple[PolicyBehavior | None, list[sim.Sensor]]:
    """
    Load behavior and sensor previously saved in a directory
    using :py:func:`export_policy_as_behavior`.

    :param      path:  The directory path

    :returns:   A policy behavior and a sensor, set to None if the could not be loaded.
    """
    path = pl.Path(path)
    behavior: PolicyBehavior | None = None
    sensors: list[sim.Sensor] = []
    with chdir(path):
        with open('behavior.yaml') as f:
            b = core.Behavior.load(f.read())
            if isinstance(b, PolicyBehavior):
                behavior = b
        with open('sensor.yaml') as f:
            sensors_as_dict = yaml.safe_load(f.read())
            sensors = [
                se for se in (sim.Sensor.load(yaml.dump(d))
                              for d in sensors_as_dict) if se
            ]
    return behavior, sensors



from __future__ import annotations

import copy
from collections.abc import Callable, Collection
from functools import partial
from typing import Any, cast

from navground import core, sim

from ..config import (DefaultObservationConfig, GroupConfig,
                      ModulationActionConfig, ObservationConfig)
from ..internal.base_env import NavgroundBaseEnv
from ..internal.group import GymAgent, create_agents_in_groups
from ..probes import RewardProbe


def update_state(sensors: list[sim.Sensor], world: sim.World,
                 agent: sim.Agent) -> Callable[[core.SensingState], None]:

    def f(state: core.SensingState) -> None:
        for sensor in sensors:
            sensor.update(agent, world, state)

    return f


class PolicyModulation(core.BehaviorModulation):

    def __init__(
            self,
            action_config: ModulationActionConfig = ModulationActionConfig(),
            observation_config: ObservationConfig = DefaultObservationConfig(),
            policy: Any = None,
            observation_fn: Callable[[core.SensingState], None] | None = None,
            deterministic: bool = True):
        super().__init__()
        self.action_config = action_config
        self.observation_config = observation_config
        self.gym_agent: GymAgent | None = None
        self.policy = policy
        self.observation_fn = observation_fn
        self.state = core.SensingState()
        self.deterministic = deterministic
        self._old_params: dict[str, Any] = {}

    def pre(self, behavior: core.Behavior, time_step: float) -> None:
        if self.observation_fn:
            self.observation_fn(self.state)
        if not self.gym_agent:
            self.gym_agent = GymAgent(action=self.action_config,
                                      observation=self.observation_config,
                                      state=self.state,
                                      behavior=behavior)
        obs = self.gym_agent.update_observation()
        act, _ = self.policy.predict(obs, deterministic=self.deterministic)
        ac = cast("ModulationActionConfig", self.gym_agent.action_config)
        params = ac.get_params_from_action(act)
        for k, v in params.items():
            self._old_params[k] = getattr(behavior, k)
            setattr(behavior, k, v)
        behavior.modulated_params = params  # type: ignore
        behavior.modulation_input = obs  # type: ignore

    def post(self, behavior: core.Behavior, time_step: float,
             cmd: core.Twist2) -> core.Twist2:
        for k, v in self._old_params.items():
            setattr(behavior, k, v)
        return cmd


def add_modulation(groups: Collection[GroupConfig],
                   policy: Any) -> Callable[[sim.World, int | None], None]:

    def f(world: sim.World, seed: int | None = None) -> None:
        for group in groups:
            for agent in group.indices.sub_sequence(world.agents):
                obs = update_state(group.get_sensors(), world, agent)
                if group.action and isinstance(
                        group.action,
                        ModulationActionConfig) and group.observation:
                    agent.behavior.add_modulation(  # type: ignore
                        PolicyModulation(group.action, group.observation,
                                         policy, obs))

    return f


# Not safe: assumes all agents have the same modulation
class ModProbe(sim.RecordProbe):

    dtype = float

    def __init__(self, ds: sim.Dataset, groups: Collection[GroupConfig]):
        super().__init__(ds)
        self._groups = groups
        self.sizes: dict[int, int] = {}

    def prepare(self, run: sim.ExperimentalRun) -> None:
        agents = create_agents_in_groups(run.world, self._groups)
        for index, agent in agents.items():
            if agent.navground and agent.navground.behavior:
                self.sizes[index] = sum([
                    len(mod.action_config.params)
                    for mod in agent.navground.behavior.modulations
                    if isinstance(mod, PolicyModulation)
                ])
        if not len(set(self.sizes.values())) < 2:
            raise ValueError(f"Unnormal param sizes {self.sizes}")
        super().prepare(run)

    def update(self, run: sim.ExperimentalRun) -> None:

        for agent in run.world.agents:
            if agent.behavior and hasattr(agent.behavior, "modulated_params"):
                params = agent.behavior.modulated_params
                for v in params.values():
                    self.data.push(v)

    def get_shape(self, world: sim.World) -> list[int]:
        sizes = list(self.sizes.values())
        if sizes:
            size = sizes[0]
        else:
            size = 0
        return [len(self.sizes), size]


def make_experiment_with_env(
    env: NavgroundBaseEnv,
    policy: Any = None,
    record_modulations: bool = False,
    terminate_when_all_idle_or_stuck: bool = False
    # record_obervations: bool = False
) -> sim.Experiment:
    exp = sim.Experiment()
    if env._scenario:
        exp.scenario = copy.deepcopy(env._scenario)
    else:
        raise ValueError("No scenario")
    if policy:
        exp.scenario.add_init(add_modulation(env.groups_config, policy))
        if record_modulations:
            exp.add_record_probe("modulation",
                                 partial(ModProbe, groups=env.groups_config))
        # if record_obervations:
        #     exp.add_record_probe("observation",
        #                          partial(ObsProbe, config=env.config))
    exp.add_record_probe("reward",
                         partial(RewardProbe, groups=env.groups_config))
    exp.terminate_when_all_idle_or_stuck = terminate_when_all_idle_or_stuck
    return exp



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any

import gymnasium as gym
import numpy as np
from pettingzoo.utils.wrappers import BaseParallelWrapper

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv

from ..internal.base_env import ResetReturn, StepReturn
from ..types import Action, Array, Observation


def mask_array_or_dict(
    value: dict[str, Array] | Array,
    indices: Iterable[int] = tuple(),
    keys: Iterable[str] = tuple()
) -> Observation:
    if isinstance(value, Mapping):
        keys = set(keys)
        return {k: v for k, v in value.items() if k not in keys}
    indices = list(indices)
    if indices:
        return np.delete(value, indices, axis=-1)
    return value


def unmask_array(values: Array,
                 indices: Iterable[int],
                 value: float = 0) -> Array:
    for i in indices:
        values = np.insert(values, i, value, axis=-1)
    return values


def mask_space(
    value: gym.spaces.Box | gym.spaces.Dict,
    indices: Iterable[int] = tuple(),
    keys: Iterable[str] = tuple()
) -> gym.spaces.Box | gym.spaces.Dict:
    if isinstance(value, Mapping):
        keys = set(keys)
        return gym.spaces.Dict({
            k: v
            for k, v in value.items() if k not in keys
        })
    indices = list(indices)
    return gym.spaces.Box(np.delete(value.low, indices),
                          np.delete(value.high, indices))


class MaskWrapper(BaseParallelWrapper):
    """
    This wrapper masks some of the observations and/or actions

    :param env: The wrapped environment
    :param observation_indices: Which observation indices to keep (if array)
    :param observation_keys: Which observation keys to keep (if dict)
    :param action_indices: Which action indices to keep (if array)
    """

    def __init__(
        self,
        env: ParallelEnv,
        observation_indices: Iterable[int] = tuple(),
        observation_keys: Iterable[str] = tuple(),
        action_indices: Iterable[int] = tuple()
    ) -> None:
        super().__init__(env)
        self._action_indices = action_indices
        self._observation_indices = observation_indices
        self._observation_keys = observation_keys

    def observation_space(self, agent):
        return mask_space(self.env.observation_space(agent),
                          indices=self._observation_indices,
                          keys=self._observation_keys)

    def action_space(self, agent):
        return mask_space(self.env.action_space(agent),
                          indices=self._action_indices)

    def reset(self,
              seed: int | None = None,
              options: dict[str, Any] | None = None) -> ResetReturn:
        obs, infos = self.env.reset(seed=seed, options=options)
        obs = {
            agent:
            mask_array_or_dict(o,
                               indices=self._observation_indices,
                               keys=self._observation_keys)
            for agent, o in obs.items()
        }
        return obs, infos

    def step(self, action: dict[int, Action]) -> StepReturn:
        action = {
            agent: unmask_array(act, indices=self._action_indices)
            for agent, act in action.items()
        }
        obs, reward, terminated, truncated, infos = self.env.step(action)
        obs = {
            agent:
            mask_array_or_dict(o,
                               indices=self._observation_indices,
                               keys=self._observation_keys)
            for agent, o in obs.items()
        }
        return obs, reward, terminated, truncated, infos



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import numpy as np
from imitation.data import types
from pettingzoo.utils.wrappers.base_parallel import BaseParallelWrapper

from ..internal.base_env import ResetReturn, StepReturn
from ..types import Action, Observation

if TYPE_CHECKING:
    from pettingzoo.utils.env import ParallelEnv

ObsType = types.DictObs | np.typing.NDArray[Any]

EnvKwargs = Mapping[str, Any]


class RolloutInfoWrapper(BaseParallelWrapper[int, Observation, Action]):

    def __init__(self, env: ParallelEnv[int, Observation, Action]):
        super().__init__(env)
        self._obs: dict[int, list[types.DictObs]
                        | list[np.typing.NDArray[Any]]] = {}
        self._rews: dict[int, list[float]] = {}

    def reset(self,
              seed: int | None = None,
              options: dict[str, Any] | None = None) -> ResetReturn:
        obss, infos = self.env.reset(seed=seed, options=options)
        self._obs = {
            i: [types.maybe_wrap_in_dictobs(obs)]  # type: ignore[misc]
            for i, obs in obss.items()
        }
        self._rews = {i: [] for i in obss}
        return obss, infos

    def step(self, actions: dict[int, Action]) -> StepReturn:
        obss, rews, terms, trucs, infos = self.env.step(actions)
        for i, obs in obss.items():
            self._obs[i].append(
                types.maybe_wrap_in_dictobs(obs))  # type: ignore[arg-type]
        for i, rew in rews.items():
            self._rews[i].append(rew)

        for i, term in terms.items():
            done = term or trucs[i]
            if done:
                info = infos[i]
                assert "rollout" not in info
                info["rollout"] = {
                    "obs": types.stack_maybe_dictobs(
                        self._obs[i]),  # type: ignore[misc]
                    "rews": np.stack(self._rews[i]),
                }
        return obss, rews, terms, trucs, infos



from __future__ import annotations

from typing import Any

import gymnasium as gym
import torch
from torch.distributions import transforms as D
from torch.distributions.categorical import Categorical
from torchrl.modules import ProbabilisticActor  # type: ignore
from torchrl.modules.distributions.continuous import TanhNormal  # type: ignore
from torchrl.modules.tensordict_module import SafeProbabilisticModule  # type: ignore

from benchmarl.models.mlp import Mlp  # type: ignore[import-not-found]

from ...types import (Action, Array, EpisodeStart, Info, Observation,
                      PyTorchObs, State)
from .mlp import MlpPolicy
from .split_mlp import SplitMlp, SplitMlpPolicy


def _has_batch_dim(observation: Array, flat_dims: int) -> bool:
    return len(observation.shape) == 2 or observation.shape[0] != flat_dims

class SingleAgentPolicy(torch.nn.Module):
    """
    This class conforms to :py:class:`navground.learning.types.PyTorchPolicy`
    and is construction from a TorchRL policy.
    """

    def __init__(self, observation_space: gym.Space,
                 action_space: gym.spaces.Box, policy: Any):
        """
        Constructs a new instance.

        :param      observation_space:  The observation space
        :param      action_space:       The action space
        :param      policy:             The TorchRL policy
        """
        super().__init__()
        self.observation_space = observation_space
        self.action_space = action_space
        # actor = policy.module[0]
        actor = policy
        dist_module = actor[1]
        if isinstance(dist_module, SafeProbabilisticModule):
            if dist_module.distribution_class in (TanhNormal, ):
                min_act = dist_module.distribution_kwargs['low'][0]
                max_act = dist_module.distribution_kwargs['high'][0]
                t = D.AffineTransform(loc=(max_act + min_act) / 2,
                                      scale=(max_act - min_act) / 2)
                self._tail = lambda x: t(x.tanh())
                # self._tail = lambda x: t(dist_module(x))
            elif dist_module.distribution_class in (Categorical, ):
                self._tail = lambda x: dist_module(x)[0]
            else:
                self._tail = lambda x: dist_module(x)
        else:
            self._tail = lambda x: dist_module(x)[0]
        self._npe = None
        if isinstance(actor, ProbabilisticActor):
            if isinstance(action_space, gym.spaces.Box):
                model = actor.module[0][0]
                self._npe = actor[0][1].module
            else:
                model = actor.module[0]
        else:
            model = actor.module[0]
        self.model = self._copy_model(model)
        if isinstance(observation_space, gym.spaces.Dict):
            self._uses_dict = True
            self._flat_dims = {
                k: gym.spaces.utils.flatdim(v)
                for k, v in observation_space.items()
            }
        else:
            self._uses_dict = False
            self._flat_dim = gym.spaces.utils.flatdim(observation_space)

    def forward(self,
                observation: PyTorchObs,
                deterministic: bool = True) -> torch.Tensor:
        logits = self.model.forward(observation)
        if self._npe:
            loc, _ = self._npe(logits)
        else:
            loc = logits
        return self._tail(loc)

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = False,
                info: Info | None = None) -> tuple[Action, State | None]:
        if self._uses_dict:
            if not isinstance(observation, dict):
                raise TypeError("Requires a dict observation")
            are_batches = set(
                _has_batch_dim(v, self._flat_dims[k])
                for k, v in observation.items())
            is_batch = all(are_batches)
            obs: PyTorchObs = {
                k: torch.from_numpy(v.reshape(-1, self._flat_dims[k]))
                for k, v in observation.items()
            }
        else:
            if isinstance(observation, dict):
                raise TypeError("Requires a flat observation")
            is_batch = _has_batch_dim(observation, self._flat_dim)
            obs = torch.from_numpy(observation.reshape(-1, self._flat_dim))
        with torch.no_grad():
            act = self.forward(obs,
                               deterministic=deterministic).detach().numpy()
            if not is_batch:
                act = act.flatten()
            return act, None

    def _copy_model(self, model: torch.nn.Module) -> torch.nn.Module:
        if isinstance(model, Mlp):
            return MlpPolicy(model)
        if isinstance(model, SplitMlp):
            return SplitMlpPolicy(model)
        raise TypeError(f"Model {type(model)} not supported")



from .centralized import CentralizedPadBehavior
from .distributed import DistributedPadBehavior
from .stop import StopAtPadBehavior

__all__ = [
    'CentralizedPadBehavior', 'DistributedPadBehavior', 'StopAtPadBehavior'
]



from __future__ import annotations

import warnings
from typing import Any, cast

import gymnasium as gym
import numpy as np
import re

from ..types import Action, EpisodeStart, Observation, PathLike, State


def space_for_onnx_tensor(x: Any) -> gym.spaces.Box:
    shape = [i for i in x.shape if isinstance(i, int)]
    ts = re.findall(r"tensor\((.*)\)", x.type)
    dtype: type[np.floating[Any]] | type[np.integer[Any]]
    if not ts:
        warnings.warn(f"Unknown type {x.type}", stacklevel=1)
        dtype = np.float32
    else:
        if ts[0] == "float":
            dtype = np.float32
        elif ts[0] in x.type:
            dtype = np.float64
        else:
            dtype = np.dtype(ts[0]).type
    return gym.spaces.Box(shape=shape, low=-np.inf, high=np.inf, dtype=dtype)


class OnnxPolicy:
    """
    This class implements the :py:class:`navground.learning.types.PolicyPredictor`
    protocol. It loads an onnx model andd then calls
    :py:meth:`onnxruntime.InferenceSession.run` to perform inference.

    :param  path:         The path
    :param  action_space: The action_space (if not specified,
                          it will be reconstructed from the onnx model)
    """

    def __init__(self, path: PathLike, action_space: gym.Space | None = None):
        import onnxruntime as ort  # type: ignore[import-untyped]

        options = ort.SessionOptions()
        options.intra_op_num_threads = 1
        options.execution_mode = ort.ExecutionMode.ORT_SEQUENTIAL
        options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
        self.ort_sess = ort.InferenceSession(path, options)
        self.input_dims = [len(x.shape) for x in self.ort_sess.get_inputs()]
        xs = self.ort_sess.get_inputs()
        ys = self.ort_sess.get_outputs()
        if len(xs) == 1 and xs[0].name == "observation":
            self._observation_space: gym.spaces.Box | gym.spaces.Dict = space_for_onnx_tensor(
                xs[0])
        else:
            ds = [(x.name, space_for_onnx_tensor(x)) for x in xs]
            self._observation_space = gym.spaces.Dict(ds)
            self._keys: list[str] = [x.name for x in xs]
        # TODO(Jerome): why can it have more than one output?
        # assert len(ys) == 1, ys
        if action_space is None:
            self._action_space: gym.Space = space_for_onnx_tensor(ys[0])
        else:
            self._action_space = action_space

    # single obs -> single action, None
    # multiple obs -> multiple obs, None

    def predict(self,
                observation: Observation,
                state: State | None = None,
                episode_start: EpisodeStart | None = None,
                deterministic: bool = True) -> tuple[Action, State | None]:
        """
        Perform one inference, calling :py:meth:`onnxruntime.InferenceSession.run`.

        The observation should respect the model specifics, also exposed
        by :py:attr:`observation_space`.

        :param      observation:    The input
        :param      state:          The state (ignored)
        :param      episode_start:  The episode start (ignored)
        :param      deterministic:  Whether deterministic (ignored)

        :returns:   The output of the inference and a `None` state
        """
        if not isinstance(observation, dict):
            vectorized = len(observation.shape) == self.input_dims[0]
            if not vectorized:
                observation = observation[np.newaxis, :]
            observation = {"observation": observation}
        else:
            vectorized = all(
                len(observation[key].shape) == length for key, length in zip(
                    self._keys, self.input_dims, strict=True))
            if not vectorized:
                observation = {
                    k: v[np.newaxis, :]
                    for k, v in observation.items()
                }
        action = self.ort_sess.run(None, observation)[0]
        if not vectorized:
            action = action[0]
        return cast("Action", action), None

    @property
    def observation_space(self) -> gym.spaces.Box | gym.spaces.Dict:
        """
        The observation space
        """
        return self._observation_space

    @property
    def action_space(self) -> gym.Space:
        """
        The action space
        """
        return self._action_space



from __future__ import annotations

import dataclasses as dc
import math
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import core, sim
from navground.core import FloatType

from ..types import Array
from .base import DataclassConfig, StateConfig

# TODO(Jerome): Complete with target state


@dc.dataclass(repr=False)
class DefaultStateConfig(DataclassConfig, StateConfig,
                         register_name="Default"):
    """
    This class configures which information from the behavior state to
    include in the global environement state.

    :param include_position: Whether observations include the position

    :param include_orientation: Whether observations include the orientation

    :param include_radius: Whether to include the radius in the observations.

    :param max_radius: Maximal radius.

    :param include_velocity: Whether to include the velocity

    :param include_angular_speed: Whether to include the angular_speed

    :param include_x: Whether observations include the x-coordinates

    :param include_y: Whether observations include the y-coordinates

    :param include_target_distance: Whether to include the target distance in the observations.

    :param max_target_distance: Maximal target distance

    :param include_target_distance_validity: Whether to include whether the target
                                             distance is valid in the observations.

    :param include_target_direction: Whether to include the target direction in the observations.

    :param include_target_direction_validity: Whether to include whether the target direction
                                              is valid in the observations.

    :param include_target_speed: Whether to include the target speed in the observations.

    :param include_target_angular_speed: Whether to include the target angular speed
                                         in the observations.
    """

    include_position: bool = False
    """Whether observations include the position."""
    include_orientation: bool = False
    """Whether observations include the orientation."""
    include_radius: bool = False
    """Whether observations include the radius."""
    max_radius: float = math.inf
    """Maximal radius."""
    include_velocity: bool = False
    """Whether to include the velocity."""
    include_angular_speed: bool = False
    """Whether to include the angular_speed."""
    include_x: bool = True
    """Whether observations include the x-coordinates"""
    include_y: bool = True
    """Whether observations include the y-coordinates"""
    use_absolute_frame: bool = True
    """Whether to use the absolute frame for velocity and target direction"""
    include_target_orientation: bool = False
    """TODO"""
    include_target_orientation_validity: bool = False
    """TODO"""
    include_target_distance: bool = False
    """Whether observations include the target direction."""
    max_target_distance: float = math.inf
    """Maximal target distance"""
    include_target_distance_validity: bool = False
    """Whether observations include the validity of the target direction."""
    include_target_direction: bool = False
    """Whether observations include the target direction."""
    include_target_direction_validity: bool = False
    """Whether observations include whether the validity of the target direction."""
    include_target_speed: bool = False
    """Whether observations include the target speed."""
    include_target_angular_speed: bool = False
    """Whether observations include the target angular speed."""
    include_all: dc.InitVar[bool] = False
    """Whether to include all fields"""

    def __post_init__(self, include_all: bool) -> None:
        super().__post_init__()
        if include_all:
            self.include_position = True
            self.include_orientation = True
            self.include_radius = True
            self.include_velocity = True
            self.include_angular_speed = True
            self.include_target_distance = True
            self.include_target_direction = True
            self.include_target_orientation = True
            self.include_target_speed = True
        self._dict_space: gym.spaces.Dict = gym.spaces.Dict()

    def get_space(self, world: sim.World) -> gym.spaces.Box:
        return cast('gym.spaces.Box',
                    gym.spaces.flatten_space(self._get_dict_space(world)))

    def get_state(self, world: sim.World) -> Array:
        rs = self._get_dict_state(world)
        vs = cast('Array', gym.spaces.flatten(self._get_dict_space(world), rs))
        if self._dtype:
            vs = vs.astype(self._dtype)
        return vs

    def _get_dict_space(self, world: sim.World) -> gym.spaces.Dict:
        if not self._dict_space:
            self._dict_space = self._make_dict_space(world)
        return self._dict_space

    def _make_dict_space(self, world: sim.World) -> gym.spaces.Dict:
        ds: dict[str, gym.Space[Any]] = {}
        num = len(world.agents)
        if self.include_velocity or self.include_target_speed:
            max_speeds = np.array([[agent.kinematics.max_speed]
                                   for agent in world.agents
                                   if agent.kinematics])
            if self.include_velocity:
                if self.use_absolute_frame:
                    max_vels = np.stack([max_speeds, max_speeds])
                    if not self.include_x:
                        max_vels = max_vels[1:]
                    if not self.include_y:
                        max_vels = max_vels[:1]
                else:
                    max_vels = np.array([[agent.kinematics.max_speed] *
                                         (agent.kinematics.dof() - 1)
                                         for agent in world.agents
                                         if agent.kinematics]).flatten()
                ds['velocity'] = gym.spaces.Box(-max_vels,
                                                max_vels,
                                                dtype=FloatType)
            if self.include_target_speed:
                ds['target_speed'] = gym.spaces.Box(-max_speeds, max_speeds)
        if self.include_angular_speed or self.include_target_angular_speed:
            max_angular_speeds = np.array([
                agent.kinematics.max_angular_speed for agent in world.agents
                if agent.kinematics
            ])
            space = gym.spaces.Box(-max_angular_speeds,
                                   max_angular_speeds,
                                   dtype=FloatType)
            if self.include_angular_speed:
                ds['angular_speed'] = space
            if self.include_target_angular_speed:
                ds['target_angular_speed'] = space
        if self.include_position:
            bl = world.bounding_box.p1
            tr = world.bounding_box.p2
            if not self.include_x:
                bl = bl[1:]
                tr = tr[1:]
            if not self.include_y:
                bl = bl[:1]
                tr = tr[:1]
            # TODO: use np.repeat
            bls = np.array([bl for _ in world.agents])
            trs = np.array([tr for _ in world.agents])
            ds['position'] = gym.spaces.Box(bls, trs, dtype=FloatType)
        if self.include_orientation:
            ds['orientation'] = gym.spaces.Box(-1,
                                               1, (num, 2),
                                               dtype=FloatType)
        if self.include_radius:
            if not math.isfinite(self.max_radius):
                self.max_radius = max(agent.radius for agent in world.agents)
            ds['radius'] = gym.spaces.Box(0,
                                          self.max_radius, (num, ),
                                          dtype=FloatType)
        if self.include_target_direction:
            dims = self._dims() if self.use_absolute_frame else 2
            if dims == 0:
                self.include_target_direction = False
            else:
                ds['target_direction'] = gym.spaces.Box(-1,
                                                        1, (num, dims),
                                                        dtype=FloatType)
                if self.include_target_direction_validity:
                    ds['target_direction_valid'] = gym.spaces.Box(
                        0, 1, (num, ), dtype=np.uint8)
        if self.include_target_distance:
            self.max_target_distance = min(
                self.max_target_distance,
                float(
                    np.linalg.norm(world.bounding_box.p1 -
                                   world.bounding_box.p2)))
            ds['target_distance'] = gym.spaces.Box(0,
                                                   self.max_target_distance,
                                                   (1, ),
                                                   dtype=FloatType)
            if self.include_target_distance_validity:
                ds['target_distance_valid'] = gym.spaces.Box(0,
                                                             1, (num, ),
                                                             dtype=np.uint8)
        if self.include_target_orientation:
            dims = 2
            if dims == 0:
                self.include_target_orientation = False
            else:
                ds['target_orientation'] = gym.spaces.Box(-1,
                                                          1, (num, dims),
                                                          dtype=FloatType)
                if self.include_target_orientation_validity:
                    ds['target_orientation_valid'] = gym.spaces.Box(
                        0, 1, (1, ), dtype=np.uint8)
        return gym.spaces.Dict(ds)

    def _get_dict_state(self, world: sim.World) -> dict[str, Array]:
        rs: dict[str, Array] = {}
        if self.include_position:
            rs['position'] = np.array([
                self._get_components(agent.position) for agent in world.agents
            ])
        if self.include_orientation:
            theta = np.array([agent.orientation for agent in world.agents])
            rs['orientation'] = np.concatenate(
                [np.cos(theta), np.sin(theta)], axis=0)
        if self.include_radius:
            rs['radius'] = np.array([agent.radius for agent in world.agents])
        if self.include_velocity:
            if self.use_absolute_frame:
                rs['velocity'] = np.array([
                    self._get_components(agent.velocity)
                    for agent in world.agents if agent.kinematics
                ])
            else:
                rs['velocity'] = np.concatenate([
                    agent.twist.relative(
                        agent.pose).velocity[:(agent.kinematics.dof() - 1)]
                    for agent in world.agents if agent.kinematics
                ])
        if self.include_angular_speed:
            rs['angular_speed'] = np.array(
                [agent.twist.angular_speed for agent in world.agents])
        if self.include_target_distance:
            values: list[float] = []
            validity: list[bool] = []
            for agent in world.agents:
                if agent.behavior:
                    distance = agent.behavior.get_target_distance()
                    values.append(min(distance or 0, self.max_target_distance))
                    validity.append(distance is not None)
                else:
                    values.append(0)
                    validity.append(False)
            rs['target_distance'] = np.array(values, dtype=FloatType)
            if self.include_target_distance_validity:
                rs['target_distance_valid'] = np.array(validity, np.uint8)
        if self.include_target_orientation:
            v_values: list[core.Vector2] = []
            validity = []
            for agent in world.agents:
                if agent.behavior:
                    orientation = agent.behavior.get_target_orientation(
                        frame=core.Frame.absolute if self.
                        use_absolute_frame else core.Frame.relative)
                    v_values.append(
                        core.unit(orientation) if orientation else np.
                        zeros(2, dtype=FloatType))
                    validity.append(orientation is not None)
                else:
                    v_values.append(np.zeros(2, dtype=FloatType))
                    validity.append(False)
            rs['target_orientation'] = np.array(v_values, dtype=FloatType)
            rs['target_orientation_valid'] = np.array(validity, np.uint8)
        if self.include_target_direction:
            v_values = []
            validity = []
            for agent in world.agents:
                if agent.behavior:
                    direction = agent.behavior.get_target_direction(
                        frame=core.Frame.absolute if self.
                        use_absolute_frame else core.Frame.relative)
                    v = direction if direction is not None else np.zeros(
                        2, dtype=FloatType)
                    if self.use_absolute_frame:
                        v = self._get_components(v)
                    v_values.append(v)
                else:
                    v_values.append(np.zeros(2, dtype=FloatType))
                    validity.append(False)
            rs['target_direction'] = np.array(v_values, dtype=FloatType)
            rs['target_direction_valid'] = np.array(validity, np.uint8)
        if self.include_target_speed:
            values = []
            for agent in world.agents:
                if agent.behavior:
                    values.append(agent.behavior.get_target_speed())
                else:
                    values.append(0)
            rs['target_speed'] = np.array(values, dtype=FloatType)
        if self.include_target_angular_speed:
            values = []
            for agent in world.agents:
                if agent.behavior:
                    values.append(agent.behavior.get_target_angular_speed())
                else:
                    values.append(0)
            rs['target_angular_speed'] = np.array(values, dtype=FloatType)
        return rs

    def _get_components(self, vs: core.Vector2) -> Array:
        if not self.include_x:
            vs = vs[1:]
        if not self.include_y:
            vs = vs[:1]
        return vs

    def _dims(self) -> int:
        return int(self.include_x) + int(self.include_y)



from __future__ import annotations

import dataclasses as dc
from typing import Any, cast

import gymnasium as gym
import numpy as np
from navground import core

from ..types import Action
from .control_action import ControlActionConfig


def to_discrete(value: float, threshold: float = 0) -> tuple[int, int]:
    if value > threshold:
        return (0, 1)
    if value < threshold:
        return (1, 0)
    return (0, 0)


def to_continuous(values: tuple[int, int]) -> float:
    if values == (1, 0):
        return 0
    if values == (0, 1):
        return 1
    return 0.5


@dc.dataclass
class BinaryControlActionConfig(ControlActionConfig,
                                register_name="BinaryControl"):
    """
    Configuration of the conversion between *discrete* control actions
    and control commands. Actions are either command accelerations
    or command velocities, depending on :py:attr:`ControlActionConfig.use_acceleration_action`.

    Discrete actions are first converted to continuous action and then,
    by the super class :py:class:`ControlActionConfig`, to control commands.

    The action space :py:class:`gymnasium.spaces.MultiBinary` is discrete,
    with two values for each continuous dimension:
    the first bit refers to moving/accelerating backwards (yes/no),
    and the second bit to moving/accelerating forwards (yes/no).

    Discrete values are encoded using a 2-bits per dimension, where

    .. list-table:: Title
       :header-rows: 1

       * - Discrete
         - Continuous
       * - [1, 0]
         - ``space.low``
       * - [0, 0]
         - 0
       * - [1, 1]
         - 0
       * - [0, 1]
         - ``space.high``

    :param dtype: The data type

    :param dof: The number of degrees of freedom of the agent

    :param max_speed: The upper bound of the speed.

    :param max_angular_speed: The upper bound of the angular speed.

    :param max_acceleration: The upper bound of the acceleration.

    :param max_angular_acceleration: The upper bound of the angular acceleration.

    :param use_acceleration_action: Whether actions are accelerations.
                                    If not set, actions are velocities.

    :param use_wheels: Whether action uses wheel speeds/acceleration
                       instead of body speeds/acceleration.
                       Only effective if the b behavior has a wheeled kinematics.

    :param fix_orientation: Whether to force the agent not to control orientation,
                            i.e., to not include the angular command in actions.

    :param has_wheels: Whether the agent as wheels.
                       If None, it will defer to the agent kinematics.
    """

    def configure(self, behavior: core.Behavior) -> None:
        super().configure(behavior)
        self._cont_space = cast('gym.spaces.Box', super().space)

    def get_cmd_from_action(self, action: Action,
                            behavior: core.Behavior | None,
                            time_step: float) -> core.Twist2:
        low = self._cont_space.low
        high = self._cont_space.high
        act = action.reshape((-1, 2)).astype(self._cont_space.dtype)
        act = ((act[..., 1] - act[..., 0]) / 2 + 0.5) * (high - low) + low
        return super().get_cmd_from_action(act, behavior, time_step)

    def get_action(self, behavior: core.Behavior, time_step: float) -> Action:
        act = ControlActionConfig.get_action(self, behavior, time_step)
        ths = (self._cont_space.high + self._cont_space.low) / 2
        d_act = np.array(
            [to_discrete(x, th) for x, th in zip(act, ths, strict=True)],
            dtype=np.int8).flatten()
        return d_act

    @property
    def space(self) -> gym.Space[Any]:
        """
        The action space.

        brake Y/N x accelerate Y/N
        """
        return gym.spaces.MultiBinary(self.action_size * 2)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from navground import sim
from pettingzoo.utils.env import ParallelEnv

from ..config import GroupConfig
from ..env import BaseEnv
from ..indices import Indices, IndicesLike
from ..parallel_env import BaseParallelEnv, MultiAgentNavgroundEnv
from ..types import Reward
from .evaluate_policies import evaluate_policies
from .evaluate_policy import evaluate_policy
from .experiment import make_experiment, make_experiment_with_env

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv


def evaluate(
    env: BaseEnv | BaseParallelEnv | VecEnv,
    n_eval_episodes: int = 10,
    deterministic: bool = True,
    return_episode_rewards: bool = False,
    warn: bool = True,
    indices: IndicesLike = Indices.all(),
) -> tuple[float, float] | tuple[list[float], list[int]]:
    """
    Similar interface as StableBaseline3
    :py:func:`stable_baselines3.common.evaluation.evaluate_policy`
    to evaluate the navground policy/behavior.

    Internally, it runs :py:func:`navground.learning.evaluation.evaluate_policy`
    or :py:func:`navground.learning.evaluation.evaluate_policies`,
    depending if ``env`` is a single or a multi-agent environment.

    :param      env:                     The environment
    :param      n_eval_episodes:         The number of episodes
    :param      deterministic:           Whether the policy is applied deterministically
    :param      return_episode_rewards:  Whether to return all episodes (vs averaging them)
    :param      warn:                    Whether to enable warnings
    :param      indices:                 The indices of the agents whose
                                         reward we want to record

    :returns:   Same as :py:func:`stable_baselines3.common.evaluation.evaluate_policy` return,
       a tuple of (mean, std dev) or a tuple
       [reward_ep_1, reward_ep_2, ...], [length_ep_1, length_ep_2, ...]
    """
    from stable_baselines3.common.vec_env import VecEnv

    if isinstance(env, ParallelEnv):
        # TODO(Jerome): make it general for heterogeneous envs
        model = cast("MultiAgentNavgroundEnv",
                     env).get_policy(env.possible_agents[0])
        models = [(indices, model)]
        result = evaluate_policies(
            env=env,
            models=models,
            n_eval_episodes=n_eval_episodes,
            deterministic=deterministic,
            return_episode_rewards=return_episode_rewards,
            warn=warn)
        if not return_episode_rewards:
            means, stds = cast("tuple[list[float], list[float]]", result)
            return means[0], stds[0]
        grs, ls, gns = cast(
            "tuple[list[list[float]], list[int], list[list[int]]]", result)
        return [value / n for value, n in zip(grs[0], gns[0], strict=True)], ls
    if isinstance(env, VecEnv):
        model = env.get_attr("policy", [0])[0]
    else:
        model = env.get_wrapper_attr("policy")
    return evaluate_policy(env=env,
                           model=model,
                           n_eval_episodes=n_eval_episodes,
                           deterministic=deterministic,
                           return_episode_rewards=return_episode_rewards,
                           warn=warn)


def compute_stats(
    experiment: sim.Experiment,
    return_episode_rewards: bool = False
) -> tuple[float, float] | tuple[list[float], list[int]]:
    rs = [
        np.asarray(run.get_record("reward"))
        for run in experiment.runs.values()
    ]
    rewards = np.concatenate([np.sum(r, axis=0) for r in rs])
    lengths = np.concatenate([[r.shape[0]] * r.shape[1] for r in rs])
    if not return_episode_rewards:
        return np.mean(rs), np.std(rs)
    return rewards.tolist(), lengths.tolist()


# returns an array of the sum of agent rewards during runs


def evaluate_with_experiment(
    scenario: sim.Scenario,
    reward: Reward,
    n_eval_episodes: int = 10,
    return_episode_rewards: bool = False,
    time_step: float = 0.1,
    steps: int = 100,
    indices: IndicesLike = Indices.all()
) -> tuple[float, float] | tuple[list[float], list[int]]:
    """
    Evaluate the navground policy/behavior using a navground experiment
    by passing an empty policy to
    :py:func:`navground.learning.evaluation.make_experiment`.

    Arguments ``n_eval_episodes`` and ``return_episode_rewards``
    mimic :py:func:`evaluate` but the return of this function
    is slighty different when ``return_episode_rewards`` is not set:
    it returns the cumulated reward over each episodes for *each agent*,
    not averaging over the group.

    :param      scenario:                The scenario
    :param      reward:                  The reward to record
    :param      n_eval_episodes:         The number of episodes
    :param      return_episode_rewards:  Whether to return all rewards (vs averaging them)
    :param      time_step:               The time step
    :param      steps:                   The steps
    :param      indices:                 The indices of the agents whose
                                         reward we want to record

    :returns:  A tuple of (mean, std dev) or a tuple
       [reward_agent_1_ep_1, reward_agent_2_ep_1, ..., reward_agent_n_ep_m],
       [length_ep_1, length_ep_2, ...]
    """
    group = GroupConfig(reward=reward, indices=Indices(indices))
    exp = make_experiment(scenario=scenario, groups=[group])
    exp.number_of_runs = n_eval_episodes
    exp.time_step = time_step
    exp.steps = steps
    exp.run()
    return compute_stats(exp, return_episode_rewards)


# Mimic SB3 `evaluate_policy`
# https://stable-baselines3.readthedocs.io/en/master/common/evaluation.html#module-stable_baselines3.common.evaluation


def evaluate_with_experiment_and_env(
    env: BaseEnv | BaseParallelEnv | VecEnv,
    n_eval_episodes: int = 10,
    return_episode_rewards: bool = False,
    indices: IndicesLike = Indices.all()
) -> tuple[list[float], list[int]] | tuple[float, float]:
    """
    Similat to :py:func:`evaluate_with_experiment` but with
    the configuration stored in the environment.

    :param      env:                     The environment
    :param      n_eval_episodes:         The number of episodes
    :param      return_episode_rewards:  Whether to return all rewards (vs averaging them)
    :param      indices:                 The indices of the agents whose
                                         reward we want to record

    :returns:   Same as :py:func:`evaluate_with_experiment`
    """
    group = GroupConfig(indices=Indices(indices))
    exp = make_experiment_with_env(env=env, groups=[group])
    exp.number_of_runs = n_eval_episodes
    exp.run()
    return compute_stats(exp, return_episode_rewards)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from stable_baselines3.common.callbacks import BaseCallback

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv

    from ...env import BaseEnv
    from ...parallel_env import BaseParallelEnv
    from ...types import PathLike

from ...evaluation import make_experiment_with_env, record_run_video


class VideoCallback(BaseCallback):
    """
    A SB3 callback that makes a video of one or more runs
    of an environment.
    """

    def __init__(self,
                 env: BaseEnv | BaseParallelEnv | VecEnv,
                 save_path: PathLike | None = None,
                 number: int = 1,
                 grouped: bool = False,
                 video_format: str = 'mp4',
                 policy: Any = None,
                 **kwargs: Any) -> None:
        """
        Constructs a new instance.

        :param      env:           The environment
        :param      save_path:     Where to save the videos
        :param      number:        The number of videos per evaluation
        :param      grouped:       Whether the policy is grouped
        :param      video_format:  The video format
        :param      policy:        The policy
        :param      kwargs:        The keywords arguments passed to the renderer
        """
        super().__init__()
        self.exp = None
        self.save_path = None
        self.number = number
        self.env = env
        self.format = video_format
        self.render_kwargs = kwargs
        self.grouped = grouped
        self.policy = policy

    def _on_step(self):
        if self.save_path is None:
            self.save_path = Path(self.model.logger.get_dir()) / 'videos'
            self.save_path.mkdir(parents=True, exist_ok=True)
        if self.exp is None:
            policy = self.policy or self.model.policy
            self.exp = make_experiment_with_env(self.env,
                                                policy=policy,
                                                record_reward=False,
                                                grouped=self.grouped,
                                                record_success=False)
        for i in range(self.number):
            name = f'{self.num_timesteps}_{i}.{self.format}'
            record_run_video(self.exp,
                             path=self.save_path / name,
                             seed=i,
                             **self.render_kwargs)
        return True



from __future__ import annotations

from dataclasses import dataclass

import torch
from benchmarl.models.mlp import Mlp, MlpConfig  # type: ignore[import-not-found]
from torch import nn
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tensordict import TensorDictBase  # type: ignore


class FlattenMlp(Mlp):

    def __init__(self, **kwargs):
        self.flatten = nn.Flatten()

    def _forward(self, tensordict: TensorDictBase) -> TensorDictBase:
        # Gather in_key
        input = torch.cat(
            [self.flatten(tensordict.get(in_key)) for in_key in self.in_keys],
            dim=-1)

        # Has multi-agent input dimension
        if self.input_has_agent_dim:
            res = self.mlp.forward(input)
            if not self.output_has_agent_dim:
                # If we are here the module is centralised and parameter shared.
                # Thus the multi-agent dimension has been expanded,
                # We remove it without loss of data
                res = res[..., 0, :]

        # Does not have multi-agent input dimension
        else:
            if not self.share_params:
                res = torch.stack(
                    [net(input) for net in self.mlp],
                    dim=-2,
                )
            else:
                res = self.mlp[0](input)

        tensordict.set(self.out_key, res)
        return tensordict


@dataclass
class FlattenMlpConfig(MlpConfig):

    @staticmethod
    def associated_class():
        return FlattenMlp



from __future__ import annotations

import copy
from collections.abc import Callable
from typing import TYPE_CHECKING

import gymnasium as gym

if TYPE_CHECKING:
    from torchrl.data import CompositeSpec  # type: ignore
    from torchrl.envs import EnvBase  # type: ignore
    from benchmarl.utils import DEVICE_TYPING  # type: ignore

from torchrl.data import Composite
from torchrl.envs.libs.pettingzoo import PettingZooWrapper  # type: ignore

from benchmarl.environments import TaskClass  # type: ignore

from ...parallel_env import MultiAgentNavgroundEnv
from ...wrappers.name_wrapper import NameWrapper


class NavgroundTaskClass(TaskClass):

    def __init__(self,
                 *args,
                 env: MultiAgentNavgroundEnv,
                 eval_env: MultiAgentNavgroundEnv | None = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        if not eval_env:
            eval_env = env
        self._env_spec = env._spec
        self._eval_env_spec = eval_env._spec
        self.config['env'] = env.asdict
        self._cont = all(
            isinstance(space, gym.spaces.Box)
            for space in env._action_space.values())
        if eval_env:
            self.config['eval_env'] = eval_env.asdict

    def get_env_fun(
        self,
        num_envs: int,
        continuous_actions: bool,
        seed: int | None,
        device: DEVICE_TYPING,
    ) -> Callable[[], EnvBase]:
        config = copy.deepcopy(self.config)
        config.pop('env')
        config.pop('eval_env', None)
        return lambda: PettingZooWrapper(
            NameWrapper(MultiAgentNavgroundEnv(**self._env_spec)),
            categorical_actions=not continuous_actions,
            device=device,
            seed=seed,
            return_state=self._env_spec.get('state', False),
            **config,
        )

    def get_eval_env_fun(
        self,
        num_envs: int,
        continuous_actions: bool,
        seed: int | None,
        device: DEVICE_TYPING,
    ) -> Callable[[], EnvBase]:
        config = copy.deepcopy(self.config)
        config.pop('env')
        config.pop('eval_env', None)
        return lambda: PettingZooWrapper(
            NameWrapper(MultiAgentNavgroundEnv(**self._eval_env_spec)),
            categorical_actions=not continuous_actions,
            device=device,
            seed=seed,
            return_state=self._env_spec.get('state', False),
            **config,
        )

    # def render_callback(self, *args, **kwargs):
    #     print('render')
    #     return super().render_callback(*args, **kwargs)

    def supports_continuous_actions(self) -> bool:
        return self._cont

    def supports_discrete_actions(self) -> bool:
        return not self._cont

    # def has_state(self) -> bool:
    #     return False

    def has_render(self, env: EnvBase) -> bool:
        return env.render_mode == 'rgb_array'

    def max_steps(self, env: EnvBase) -> int:
        return int(env._env.max_duration // env._env.time_step)

    def group_map(self, env: EnvBase) -> dict[str, list[str]]:
        return env.group_map

    def state_spec(self, env: EnvBase) -> CompositeSpec | None:
        if 'state' in env.observation_spec:
            return Composite({"state": env.observation_spec["state"].clone()})
        return None

    def action_mask_spec(self, env: EnvBase) -> CompositeSpec | None:
        return None

    def observation_spec(self, env: EnvBase) -> CompositeSpec:
        observation_spec = env.observation_spec.clone()
        for group in self.group_map(env):
            group_obs_spec = observation_spec[group]
            for key in list(group_obs_spec.keys()):
                if key != "observation":
                    del group_obs_spec[key]
        if "state" in observation_spec:
            del observation_spec["state"]
        return observation_spec

    def info_spec(self, env: EnvBase) -> CompositeSpec | None:
        observation_spec = env.observation_spec.clone()
        for group in self.group_map(env):
            group_obs_spec = observation_spec[group]
            for key in list(group_obs_spec.keys()):
                if key != "info":
                    del group_obs_spec[key]
        if "state" in observation_spec:
            del observation_spec["state"]
        return observation_spec

    def action_spec(self, env: EnvBase) -> CompositeSpec:
        return env.full_action_spec

    @staticmethod
    def env_name() -> str:
        return "navground"



from __future__ import annotations

import dataclasses as dc
from collections.abc import Sequence
from functools import reduce
from typing import SupportsFloat, cast

import gymnasium as gym
import numpy as np

from ...types import Action, AnyPolicyPredictor, Array, Observation

# Could/should replicate the Torch RL Env.rolling
# that returns a dictionary of {'action', 'next', ...}


def stack_dict(ds: Sequence[dict[str, Array]]) -> dict[str, Array]:
    keys = reduce(lambda x, y: x | y, (set(d.keys()) for d in ds))
    return {key: np.array([d[key] for d in ds if key in d]) for key in keys}


@dc.dataclass
class Rollout:
    observation: Observation
    action: Action
    reward: Array
    termination: bool
    truncation: bool
    info: dict[str, Array]


def rollout(env: gym.Env,
            max_steps: int = 1000,
            policy: AnyPolicyPredictor | None | str = None,
            deterministic: bool = True,
            seed: int | None = None,
            options: dict = {}) -> Rollout:
    obss: list[Observation] = []
    infos: list[dict[str, Array]] = []
    rews: list[SupportsFloat] = []
    acts: list[Action] = []
    obs, info = env.reset(seed=seed, options=options)
    for _ in range(max_steps):
        if isinstance(policy, str):
            act = info[policy]
        elif policy is None:
            act = env.action_space.sample()
        else:
            act, _ = policy.predict(obs, deterministic=deterministic)
        acts.append(act)
        obs, rew, term, trunc, info = env.step(act)
        obss.append(obs)
        rews.append(rew)
        infos.append(info)
        if term or trunc:
            break
    if isinstance(env.observation_space, gym.spaces.Dict):
        obs = stack_dict(cast('list[dict[str, Array]]', obss))
    else:
        obs = np.asarray(obss)
    info = stack_dict(infos)
    return Rollout(observation=obs,
                   reward=np.asarray(rews),
                   termination=term,
                   truncation=trunc,
                   info=info,
                   action=np.asarray(acts))



from __future__ import annotations

import tqdm.auto


def setup_tqdm() -> None:
    """
    Disables the progress while saving the DAgger datasets
    and configures tqdm to use ``auto``.

    :returns:   { description_of_the_return_value }
    :rtype:     None
    """
    try:
        import datasets.utils  # type: ignore[import-untyped]

        datasets.utils.tqdm = tqdm.auto.tqdm
        datasets.utils.disable_progress_bar()
    except ImportError:
        pass
    try:
        import imitation.algorithms.bc

        imitation.algorithms.bc.tqdm = tqdm.auto  # type: ignore[attr-defined]
    except ImportError:
        pass



def test_import():
    import navground.learning.parallel_env  # noqa


def test_create_env():
    from navground.learning.examples.cross import get_env

    penv = get_env(multi_agent=True)
    penv.reset(seed=0)


def test_load_env():
    from navground.learning.io import load_env
    import pathlib

    path = pathlib.Path(__file__).parent / 'parallel_env.yaml'
    penv = load_env(path)
    penv.reset(seed=0)



from __future__ import annotations

import warnings
from collections.abc import Callable, Collection, Iterable, Sequence
from typing import TYPE_CHECKING

from navground import core, sim

from ..behaviors import GroupedPolicyBehavior, PolicyBehavior
from ..config import (ControlActionConfig, DefaultObservationConfig,
                      GroupConfig, merge_groups_configs)
from ..env import BaseEnv
from ..indices import Indices, IndicesLike
from ..internal.base_env import NavgroundBaseEnv
from ..parallel_env import BaseParallelEnv
from ..types import AnyPolicyPredictor, Bounds, ObservationTransform, PathLike

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv

AnyPolicyBehavior = PolicyBehavior | GroupedPolicyBehavior


def set_policy_behavior(world: sim.World,
                        indices: IndicesLike,
                        policy: AnyPolicyPredictor | PathLike,
                        action_config: ControlActionConfig,
                        observation_config: DefaultObservationConfig,
                        sensors: Sequence[sim.Sensor] = tuple(),
                        deterministic: bool = True,
                        grouped: bool = False,
                        pre: ObservationTransform | None = None) -> set[int]:
    indices = Indices(indices)
    all_indices = indices.as_set(len(world.agents))
    if grouped:
        behavior_cls: type[AnyPolicyBehavior] = GroupedPolicyBehavior
    else:
        behavior_cls = PolicyBehavior
    for i in all_indices:
        agent = world.agents[i]
        if agent.behavior:
            agent.behavior = behavior_cls.clone_behavior(  # type: ignore[attr-defined]
                behavior=agent.behavior,
                policy=policy,
                action_config=action_config,
                observation_config=observation_config,
                deterministic=deterministic,
                pre=pre)
            if sensors:
                agent.state_estimations = list(sensors)
        else:
            all_indices.remove(i)
    return set(all_indices)


def set_policy_behavior_with_env(world: sim.World,
                                 env: BaseEnv | BaseParallelEnv | VecEnv,
                                 policy: AnyPolicyPredictor | PathLike,
                                 indices: IndicesLike = "ALL",
                                 deterministic: bool = True,
                                 grouped: bool = False) -> set[int]:
    indices = Indices(indices)
    if not isinstance(env.unwrapped, NavgroundBaseEnv):
        warnings.warn(f"Environment {env} is not a Navground environment.",
                      stacklevel=1)
        return set()
    indices_set = indices.as_set(len(world.agents))
    all_indices: set[int] = set()
    for group in env.unwrapped.groups_config:
        group_indices = group.indices.as_set(len(world.agents)) & indices_set
        if group_indices and isinstance(
                group.action, ControlActionConfig) and isinstance(
                    group.observation, DefaultObservationConfig):
            set_policy_behavior(world,
                                policy=policy,
                                indices=Indices(group_indices),
                                action_config=group.action,
                                observation_config=group.observation,
                                sensors=group.get_sensors(),
                                deterministic=deterministic,
                                grouped=grouped)
            all_indices |= group_indices
    return all_indices


def to_bounds(bb: sim.BoundingBox) -> Bounds:
    return bb.p1, bb.p2


def is_outside(p: core.Vector2, bounds: Bounds) -> bool:
    return any(p < bounds[0]) or any(p > bounds[1])


def is_any_agents_outside(
    bounds: Bounds | None = None,
    agent_indices: Indices = Indices({0})
) -> Callable[[sim.World], bool]:

    def f(world: sim.World) -> bool:
        nonlocal bounds
        if bounds is None:
            bounds = to_bounds(world.bounding_box)
        return any(
            is_outside(agent.position, bounds)
            for agent in agent_indices.sub_sequence(world.agents))

    return f


def is_outside_world_bounds(
        world: sim.World,
        bounds: Bounds | None = None
) -> Callable[[sim.Agent, sim.World], bool]:
    if bounds is None:
        bounds = to_bounds(world.bounding_box)

    def f(agent: sim.Agent, world: sim.World) -> bool:
        return is_outside(agent.position, bounds)

    return f


Condition = tuple[Indices, Callable[[sim.Agent, sim.World], bool], bool | None]


def make_termination_condition(
        any_of: Iterable[Condition], all_of: Iterable[Condition],
        not_terminal: Iterable[Condition]) -> Callable[[sim.World], bool]:

    def f(world: sim.World) -> bool:
        value = False
        for indices, fn, result in any_of:
            for agent in indices.sub_sequence(world.agents):
                if fn(agent, world):
                    value = True
                    if not hasattr(agent, '_success'):
                        agent._success = result  # type: ignore[attr-defined]
        c_value = None
        for indices, fn, result in all_of:
            for agent in indices.sub_sequence(world.agents):
                if fn(agent, world):
                    if c_value is None:
                        c_value = True
                    if not hasattr(agent, '_success'):
                        agent._success = result  # type: ignore[attr-defined]
                else:
                    c_value = False
        if c_value:
            value = True
        for indices, fn, result in not_terminal:
            for agent in indices.sub_sequence(world.agents):
                if not hasattr(agent, '_success') and fn(agent, world):
                    agent._success = result  # type: ignore[attr-defined]
        return value

    return f


# pickle does not work with functions factories
# so better to write it has a callable class
class InitPolicyBehavior:
    """
    A navground scenario initializer to configure groups of agents.

    It is designed to be added to a scenario, like

    >>> from navground import sim
    >>> from navground.learning import GroupConfig
    >>>
    >>> scenario = sim.load_scenario(...)
    >>> groups = [GroupConfig(policy='policy.onnx', color='red', indices='ALL')]
    >>> scenario.add_init(InitPolicyBehavior(groups=groups))
    >>> world = scenario.make_world(seed=101)
    >>> another_world = scenario.make_world(seed=313)

    :param groups: The configuration of groups of agents
    :param bounds: Optional termination boundaries
    :param terminate_outside_bounds: Whether to terminate
        if some of the agents exits the boundaries
    :param  grouped:       Whether the policy is grouped.
    :param  deterministic: Whether to apply the policies deterministically
    :param  pre:           An optional transformation to apply to observations
    """

    def __init__(self,
                 groups: Collection[GroupConfig] = tuple(),
                 bounds: Bounds | None = None,
                 terminate_outside_bounds: bool = True,
                 deterministic: bool = True,
                 grouped: bool = False,
                 pre: ObservationTransform | None = None) -> None:
        self.groups = groups
        self.bounds = bounds
        self.terminate_outside_bounds = terminate_outside_bounds
        self.deterministic = deterministic
        self.grouped = grouped
        self.pre = pre

    def __call__(self, world: sim.World, seed: int | None = None) -> None:
        """
        Configures the policy, sensor, color, ... of the agents according
        to their group.

        :param      world:  The world
        :param      seed:   The random seed
        """
        agent_indices: set[int] = set()
        all_of: list[Condition] = []
        any_of: list[Condition] = []
        not_terminal: list[Condition] = []
        for group in self.groups:
            if group.policy and isinstance(
                    group.action, ControlActionConfig) and isinstance(
                        group.observation, DefaultObservationConfig):
                group_indices = set_policy_behavior(
                    world,
                    action_config=group.action,
                    observation_config=group.observation,
                    sensors=group.get_sensors(),
                    policy=group.policy,
                    indices=group.indices,
                    deterministic=self.deterministic,
                    grouped=self.grouped,
                    pre=self.pre)
                agent_indices |= group_indices
                for i in group_indices:
                    if group.color:
                        world.agents[i].color = group.color
                    if group.tag:
                        world.agents[i].add_tag(group.tag)
                if group.success_condition:
                    c = Indices(group_indices), group.success_condition, True
                    if group.terminate_on_success:
                        all_of.append(c)
                    else:
                        not_terminal.append(c)
                if group.failure_condition:
                    c = Indices(group_indices), group.failure_condition, False
                    if group.terminate_on_failure:
                        any_of.append(c)
                    else:
                        not_terminal.append(c)

        if agent_indices and self.terminate_outside_bounds:
            any_of.append((Indices(agent_indices),
                           is_outside_world_bounds(world, self.bounds), None))
        if all_of or any_of or not_terminal:
            tc = make_termination_condition(all_of=all_of,
                                            any_of=any_of,
                                            not_terminal=not_terminal)
            world.set_termination_condition(tc)

    @classmethod
    def with_env(
            cls,
            env: BaseEnv | BaseParallelEnv | VecEnv,
            groups: Collection[GroupConfig] = tuple(),
            bounds: Bounds | None = None,
            terminate_outside_bounds: bool | None = None,
            deterministic: bool = True,
            grouped: bool = False,
            pre: ObservationTransform | None = None) -> InitPolicyBehavior:
        """
        Returns a scenario initializer using the configuration stored in
        an environment.

        groups are merged using :py:func:`navground.learning.config.merge_groups_configs`.

        :param env:    The environment
        :param groups:      The configuration of groups of agents
        :param bounds:      Optional termination boundaries
        :param terminate_outside_bounds: Whether to terminate if some of
            the agents exits the boundaries
        :param  grouped:       Whether the policy is grouped.
        :param  deterministic: Whether to apply the policies deterministically
        :param  pre:           An optional transformation to apply to observations

        :returns:   The scenario initializer.
        """
        from stable_baselines3.common.vec_env import VecEnv

        if isinstance(env, VecEnv):
            if not bounds:
                bounds = env.get_attr('bounds')[0]
            if terminate_outside_bounds is None:
                terminate_outside_bounds = env.get_attr(
                    'truncate_outside_bounds')[0]
            env_groups = env.get_attr('groups_config')[0]
            possible_agents = env.get_attr('_possible_agents')[0]
        else:
            if not isinstance(env.unwrapped, NavgroundBaseEnv):
                raise TypeError(
                    f"Environment {env} is not a Navground environment.")
            if not bounds:
                bounds = env.unwrapped.bounds
            if terminate_outside_bounds is None:
                terminate_outside_bounds = env.unwrapped.truncate_outside_bounds
            env_groups = env.unwrapped.groups_config
            possible_agents = env.unwrapped._possible_agents
        if terminate_outside_bounds is None:
            terminate_outside_bounds = False
        groups = merge_groups_configs(groups, env_groups, len(possible_agents))
        return cls(groups, bounds, terminate_outside_bounds, deterministic,
                   grouped, pre)



from __future__ import annotations

from navground import core


class StopAtPadBehavior(core.Behavior, name="StopAtPad"):
    """
    A centralized behavior for
    :py:class:`navground.learning.scenarios.PadScenario`, that
    stops the agent before entering the pad.
    """

    def __init__(self,
                 kinematics: core.Kinematics | None = None,
                 radius: float = 0,
                 pad_width: float = 0.5,
                 tau: float = 0.5):
        """
        Constructs a new instance.

        :param      kinematics:  The kinematics
        :param      radius:      The radius
        :param      pad_width:   The pad width
        :param      tau:         The time to break
        """
        super().__init__(kinematics=kinematics, radius=radius)
        self._pad_width = pad_width
        self._tau = tau

    @property
    def pad_width(self) -> float:
        """
        The pad width

        :returns:   pad width
        """
        return self._pad_width

    @pad_width.setter
    def pad_width(self, value: float) -> None:
        """
        Sets the pad width

        :param      value:  The desired positive value
        """
        self._pad_width = max(0, value)

    @property
    def tau(self) -> float:
        """
        The braking time

        :returns:   braking time
        """
        return self._tau

    @tau.setter
    def tau(self, value: float) -> None:
        """
        Sets the braking time

        :param      value:  The desired positive value
        """
        self._tau = max(0, value)

    def cmd_twist_towards_velocity(self, velocity: core.Vector2,
                                   time_step: float) -> core.Twist2:
        if self.target.direction is None:
            speed = 0.0
        else:
            x = self.target.direction.dot(self.position)
            b = self.pad_width / 2
            if x < -b:
                speed = min(self.optimal_speed, (-b - x) / self.tau)
            else:
                speed = self.optimal_speed
        return core.Twist2((speed, 0), 0, frame=core.Frame.relative)



from __future__ import annotations

import pathlib
from collections.abc import Callable
from dataclasses import dataclass, field
from itertools import islice
from types import MethodType
from typing import TYPE_CHECKING, Any

import numpy as np
import yaml
from pettingzoo.utils.env import ParallelEnv

from navground import sim

from ..config import GroupConfig
from ..env import BaseEnv
from ..types import AnyPolicyPredictor, PathLike
from .experiment import make_experiment_with_env

if TYPE_CHECKING:
    from imitation.algorithms.bc import BCLogger
    from stable_baselines3.common.logger import Logger, TensorBoardOutputFormat
    from stable_baselines3.common.off_policy_algorithm import \
        OffPolicyAlgorithm
    from stable_baselines3.common.policies import BasePolicy
    from stable_baselines3.common.vec_env import VecEnv
    from torch.utils.tensorboard.writer import SummaryWriter

    from ..il.base import BaseILAlgorithm
    from ..parallel_env import BaseParallelEnv


def log_yaml(writer: SummaryWriter, name: str, data: dict[str, Any]) -> None:
    text = f"```yaml\n{yaml.safe_dump(data)}```"
    writer.add_text(name, text, 0)  # type: ignore[no-untyped-call]
    writer.flush()  # type: ignore[no-untyped-call]


def log_env(writer: SummaryWriter,
            env: BaseEnv | VecEnv | BaseParallelEnv) -> None:
    from stable_baselines3.common.vec_env import VecEnv

    if isinstance(env, VecEnv):
        value = env.get_attr("asdict", [0])[0]
    elif isinstance(env, ParallelEnv):
        value = env.asdict  # type: ignore
    else:
        value = env.get_wrapper_attr("asdict")
    log_yaml(writer, "environment", value)


def log_graph(writer: SummaryWriter,
              env: BaseEnv | VecEnv | BaseParallelEnv,
              policy: Any,
              index: int = 0) -> None:
    import torch as th

    if isinstance(env, ParallelEnv):
        obs = env.observation_space(0).sample()
    else:
        obs = env.observation_space.sample()
    if isinstance(obs, dict):
        x: th.Tensor | dict[str, th.Tensor] = {
            k: th.from_numpy(v[np.newaxis, :])
            for k, v in obs.values()
        }
    else:
        x = th.from_numpy(obs[np.newaxis, :])
    writer.add_graph(policy.actor, x, use_strict_trace=True,
                     verbose=False)  # type: ignore[no-untyped-call]


def get_tb_logger(logger: Logger) -> TensorBoardOutputFormat | None:
    from stable_baselines3.common.logger import TensorBoardOutputFormat
    fs = [
        f for f in logger.output_formats
        if isinstance(f, TensorBoardOutputFormat)
    ]
    if fs:
        return fs[0]
    return None


@dataclass
class VideoConfig:
    """
    The configuration to log videos
    """
    number: int = 0
    """number of runs per logging update to make a video from"""
    fps: int = 30
    """fps"""
    factor: float = 1.0
    """real-time factor for the videos"""
    color: str = ''
    """color of the agents that are learning"""
    tensorboard: bool = True
    """whether to add to the tensorboard log"""
    file: bool = True
    """whether to add to save to a file"""
    format: str = "mp4"
    """video file format"""


@dataclass
class TrajectoryPlotConfig:
    """
    The configuration to log trajectory plots
    """
    number: int = 0
    """number of runs whose trajectory to plot"""
    columns: int = 1
    """number of columns"""
    color: Callable[[sim.Agent], str] | None = None
    """color of the agents"""
    step: int = 10
    """time steps interval at which to draw the agent in the plots"""
    width: float = 10.0
    """total width of the plots"""
    world_kwargs: dict[str, Any] = field(default_factory=dict)
    """The world kwargs passed to :py:func:navground.sim.pyplot_helpers.plot_runs"""
    hide_axes: bool = True
    """whether to hide the axis"""
    tensorboard: bool = True
    """whether to add to the tensorboard log"""
    file: bool = True
    """whether to add to save to a file"""


def record_values(logger: Logger, key: str,
                  values: np.typing.NDArray[np.floating[Any]]) -> None:
    logger.record(f"eval/{key}/min", np.min(values))
    logger.record(f"eval/{key}/max", np.max(values))
    logger.record(f"eval/{key}/mean", np.mean(values))
    logger.record(f"eval/{key}/std_dev", np.std(values))


class EvalLog:

    def __init__(self,
                 video_config: VideoConfig = VideoConfig(),
                 plot_config: TrajectoryPlotConfig = TrajectoryPlotConfig(),
                 episodes: int = 100,
                 hparams: dict[str, Any] = {},
                 data: dict[str, Any] = {},
                 log_graph: bool = False,
                 reward: bool = True,
                 collisions: bool = True,
                 efficacy: bool = True,
                 safety_violation: bool = True,
                 duration: bool = False,
                 processes: int = 1,
                 use_multiprocess: bool = False,
                 use_onnx: bool = True,
                 grouped: bool = False,
                 every: int = 1):
        self.plot_config = plot_config
        self.video_config = video_config
        self.episodes = episodes
        self.hparams = hparams
        self.data = data
        self.log_graph = log_graph
        self.record_collisions = collisions
        self.record_safety_violation = safety_violation
        self.record_efficacy = efficacy
        self.record_reward = reward
        self.record_duration = duration
        self.processes = processes
        self.use_multiprocess = use_multiprocess
        self.use_onnx = use_onnx or self.processes > 1
        self.grouped = grouped
        self.every = every
        self._count = 0

    @property
    def record_pose(self) -> bool:
        return self.plot_config.number > 0 or self.video_config.number > 0

    @property
    def number_of_runs(self) -> int:
        return max(0, self.plot_config.number, self.video_config.number,
                   self.episodes)

    def init(self, env: BaseEnv | VecEnv | BaseParallelEnv, policy: BasePolicy,
             logger: Logger) -> None:
        self._init_logger(logger)
        self._init_eval_exp(env=env, policy=policy)
        if self.tb_writer:
            if self.log_graph:
                log_graph(self.tb_writer, env, policy)
            log_env(self.tb_writer, env)
            if self.data:
                for key, value in self.data.items():
                    log_yaml(self.tb_writer, key, value)
        if self.hparams:
            from stable_baselines3.common.logger import HParam

            metric_dict = {"eval/reward/mean": 0.0, "eval/reward/min": 0.0}
            self.logger.record("hparams",
                               HParam(self.hparams, metric_dict),
                               exclude=("stdout", "log", "json", "csv"))

    def _init_logger(self, logger: Logger) -> None:
        from stable_baselines3.common.logger import TensorBoardOutputFormat
        self.logger = logger
        formats = [
            fmt for fmt in logger.output_formats
            if isinstance(fmt, TensorBoardOutputFormat)
        ]
        self.tb_writer = formats[0].writer if formats else None

    def _init_eval_exp(self, env: BaseEnv | VecEnv | BaseParallelEnv,
                       policy: BasePolicy) -> None:
        if self.use_onnx:
            _policy: AnyPolicyPredictor | PathLike = "_policy.onnx"
            self._policy = policy
        else:
            _policy = policy

        group = GroupConfig(policy=_policy, color=self.video_config.color)
        exp = make_experiment_with_env(env=env,
                                       groups=[group],
                                       record_reward=self.record_reward,
                                       grouped=self.grouped)
        self.eval_exp = exp
        self.eval_exp.number_of_runs = self.number_of_runs
        self.eval_exp.record_config.pose = self.record_pose
        self.eval_exp.record_config.collisions = self.record_collisions
        self.eval_exp.record_config.efficacy = self.record_efficacy
        self.eval_exp.record_config.safety_violation = self.record_safety_violation
        if self.plot_config.number > 0 or self.video_config.number > 0:
            self.eval_exp.record_config.world = True
        # self.eval_exp.record_config.deadlocks = True

    @property
    def video_directory(self) -> pathlib.Path:
        d = self.logger.get_dir()
        if d is None:
            raise ValueError("Logging directory not set")
        path = pathlib.Path(d) / "videos"
        path.mkdir(exist_ok=True)
        return path

    @property
    def plot_directory(self) -> pathlib.Path:
        d = self.logger.get_dir()
        if d is None:
            raise ValueError("Logging directory not set")
        path = pathlib.Path(d) / "figures"
        path.mkdir(exist_ok=True)
        return path

    def evaluate(self, global_step: int) -> None:
        self._count += 1
        if self._count % self.every:
            return
        if self.eval_exp.number_of_runs:
            if self.use_onnx:
                from ..onnx import export
                export(self._policy, "_policy.onnx")

            self.eval_exp.remove_all_runs()
            if self.processes > 1:
                self.eval_exp.run_mp(number_of_processes=self.processes,
                                     keep=True,
                                     use_multiprocess=self.use_multiprocess)
            else:
                self.eval_exp.run()
        else:
            return
        if self.plot_config.number > 0:
            from matplotlib import pyplot as plt
            from stable_baselines3.common.logger import Figure

            from navground.sim.pyplot_helpers import plot_runs

            runs = list(self.eval_exp.runs.values())[:self.plot_config.number]
            fig = plot_runs(runs,
                            columns=self.plot_config.columns,
                            step=self.plot_config.step,
                            with_agent=True,
                            with_world=True,
                            color=self.plot_config.color,
                            width=self.plot_config.width,
                            world_kwargs=self.plot_config.world_kwargs,
                            hide_axes=self.plot_config.hide_axes)
            if self.plot_config.tensorboard:
                self.logger.record("eval/figure",
                                   Figure(fig, close=True),
                                   exclude=("stdout", "log", "json", "csv"))
            if self.plot_config.file:
                path = self.plot_directory / f"figure_{global_step}.pdf"
                plt.savefig(str(path))
            plt.close()

        if self.video_config.number > 0:
            import torch as th
            from stable_baselines3.common.logger import Video

            from navground.sim.ui.video import (MOVIEPY_VERSION,
                                                make_video_from_run)

            for i, run in islice(self.eval_exp.runs.items(),
                                 self.video_config.number):

                video = make_video_from_run(run,
                                            factor=self.video_config.factor)
                frames = np.array(
                    list(video.iter_frames(fps=self.video_config.fps)))
                # to (T, C, H, W)
                frames = frames.transpose(0, 3, 1, 2)
                # to (1=B, T, C, H, W)
                frames = np.expand_dims(frames, axis=0)
                if self.video_config.tensorboard:
                    self.logger.record(f"eval/video_{i}",
                                       Video(th.from_numpy(frames),
                                             fps=self.video_config.fps),
                                       exclude=("stdout", "log", "json",
                                                "csv"))
                if self.video_config.file:
                    path = (
                        self.video_directory /
                        f"video_{global_step}_{i}.{self.video_config.format}")
                    if path.suffix.lower() == ".gif":
                        video.write_gif(str(path), fps=self.video_config.fps)
                    else:
                        kwargs = {'audio': False, 'logger': None}
                        if MOVIEPY_VERSION == 1:
                            kwargs['verbose'] = False
                        video.write_videofile(str(path),
                                              fps=self.video_config.fps,
                                              **kwargs)

        steps = [run.recorded_steps for run in self.eval_exp.runs.values()]
        record_values(self.logger, "steps", np.asarray(steps))

        success = [
            np.mean(np.asarray(run.get_record("success")) > 0)
            for run in self.eval_exp.runs.values() if "success" in run.records
        ]
        if success:
            record_values(self.logger, "success", np.asarray(success))

        rewards = np.array([
            np.sum(run.get_record("reward"), axis=0)
            for run in self.eval_exp.runs.values()
        ])
        record_values(self.logger, "reward", rewards)
        if self.tb_writer:
            self.tb_writer.add_histogram(
                'eval/reward/distribution',  # type: ignore[no-untyped-call]
                values=rewards,
                global_step=global_step)
        if self.record_safety_violation:
            sv = np.array([
                np.sum(run.safety_violations, axis=0)
                for run in self.eval_exp.runs.values()
            ])
            record_values(self.logger, "safety_violations", sv)
        if self.record_efficacy:
            efficacy = np.asarray([
                np.mean(run.efficacy, axis=0)
                for run in self.eval_exp.runs.values()
            ])
            record_values(self.logger, "efficacy", efficacy)
        if self.record_collisions:
            collisions = np.asarray(
                [len(run.collisions) for run in self.eval_exp.runs.values()])
            record_values(self.logger, "collisions", collisions)
        if self.record_duration:
            duration = np.asarray(
                [run.final_sim_time for run in self.eval_exp.runs.values()])
            record_values(self.logger, "duration", duration)
        # deadlocks = np.asarray(
        #     [sum(run.deadlocks >= 0) for run in self.eval_exp.runs.values()])
        # record_values(self.logger, "deadlocks", deadlocks)


def config_eval_log(
        model: OffPolicyAlgorithm | BaseILAlgorithm,
        env: BaseEnv | VecEnv | BaseParallelEnv | None = None,
        video_config: VideoConfig = VideoConfig(),
        plot_config: TrajectoryPlotConfig = TrajectoryPlotConfig(),
        every: int = 1,
        episodes: int = 100,
        hparams: dict[str, Any] = {},
        data: dict[str, Any] = {},
        log_graph: bool = False,
        reward: bool = True,
        collisions: bool = True,
        efficacy: bool = True,
        safety_violation: bool = True,
        duration: bool = False,
        processes: int = 1,
        use_multiprocess: bool = False,
        use_onnx: bool = True,
        grouped: bool = False) -> None:
    """
    Configure the model logger to log additional data:

    - trajectory plots
    - trajectory videos
    - statistics on reward, collisions, efficacy and safety violations
    - hparams
    - data as YAML
    - model policy graph
    - a YAML representation of the environment (at the begin of the logging)

    :param      model:             The model being trained
    :param      env:               The testing environment
    :param      video_config:      The video configuration
    :param      plot_config:       The plot configuration
    :param      episodes:          The number of episodes over which to compute statistics
    :param      hparams:           The hparams
    :param      data:              The data
    :param      log_graph:         The log graph
    :param      collisions:        Whether to record episodes' collisions
    :param      efficacy:          Whether to record episodes' efficacy
    :param      safety_violation:  Whether to record episodes' safety violation
    :param      duration:          Whether to record episodes' duration
    :param      processes:         Number of processes to use
    :param      use_multiprocess:  Whether to use ``multiprocess`` instead of ``multiprocessing``
    :param      use_onnx:          Whether to use onnx for inference
    :param      grouped:           Whether the policy is grouped.
    """
    from stable_baselines3.common.off_policy_algorithm import \
        OffPolicyAlgorithm

    if env is None:
        env = model.get_env()
    assert env is not None
    log = EvalLog(video_config=video_config,
                  plot_config=plot_config,
                  episodes=episodes,
                  hparams=hparams,
                  data=data,
                  log_graph=log_graph,
                  reward=reward,
                  collisions=collisions,
                  efficacy=efficacy,
                  safety_violation=safety_violation,
                  duration=duration,
                  processes=processes,
                  use_multiprocess=use_multiprocess,
                  use_onnx=use_onnx,
                  grouped=grouped,
                  every=every)
    if isinstance(model, OffPolicyAlgorithm):
        logger = model.logger
    else:
        logger = model.logger.default_logger
    log.init(policy=model.policy, logger=logger, env=env)
    if isinstance(model, OffPolicyAlgorithm):
        _config_sb3(model, log)
    else:
        _config_il(model, log)


# In SB2.6 _dump_logs has been renamed to dump_logs
def _config_sb3(model: OffPolicyAlgorithm, log: EvalLog) -> None:
    _dump_logs = model.dump_logs
    _excluded_save_params = model._excluded_save_params

    def dump_logs(model: OffPolicyAlgorithm) -> None:
        log.evaluate(global_step=model.num_timesteps)
        _dump_logs()

    def excluded_save_params(self) -> list[str]:
        return _excluded_save_params() + ['dump_logs', '_excluded_save_params']

    model.dump_logs = MethodType(dump_logs, model)  # type: ignore
    model._excluded_save_params = MethodType(  # type: ignore[method-assign]
        excluded_save_params, model)


def _config_il(model: BaseILAlgorithm, log: EvalLog) -> None:
    if not model.logger:
        return

    from ..il import BC, DAgger

    if isinstance(model, BC):
        bc_logger = model._trainer._bc_logger
    elif isinstance(model, DAgger):
        bc_logger = model._bc_trainer._bc_logger
    else:
        return

    _log_batch = bc_logger.log_batch

    def log_batch(logger: BCLogger, batch_num: int, batch_size: int,
                  num_samples_so_far: int, *args: Any, **kwargs: Any) -> None:
        log.evaluate(global_step=logger._tensorboard_step)
        _log_batch(batch_num, batch_size, num_samples_so_far, *args, **kwargs)

    bc_logger.log_batch = MethodType(log_batch, bc_logger)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from tensordict import TensorDictBase  # type: ignore
    from torch import nn

import copy
from collections import ChainMap
from collections.abc import Collection, Sequence
from dataclasses import MISSING, dataclass, field

import torch
from torchrl.modules import MultiAgentMLP  # type: ignore

from benchmarl.models.common import Model  # type: ignore[import-not-found]
from benchmarl.models.common import ModelConfig

from ...types import PyTorchObs


def compute_input_size(input_spec, i_spec, keys):
    if i_spec is None:
        return sum([spec.shape[-1] for spec in input_spec])
    if isinstance(i_spec, slice):
        return sum(
            [len(list(range(spec.shape[-1]))[i_spec]) for spec in input_spec])
    return sum([
        spec.shape[-1] for spec, key in zip(input_spec, keys, strict=True)
        if key[-1] in i_spec
    ])

    # self.input_features =


def compute_input(tensordict, i_spec, keys):
    # print('compute_input', i_spec, keys)
    if i_spec is None:
        return torch.cat([tensordict.get(in_key) for in_key in keys], dim=-1)
    if isinstance(i_spec, slice):
        return torch.cat([tensordict.get(in_key) for in_key in keys],
                         dim=-1)[..., i_spec]
    return torch.cat(
        [tensordict.get(in_key) for in_key in keys if in_key[-1] in i_spec],
        dim=-1)


class SplitMlp(Model):

    def __init__(
        self,
        mlps_specs=[],
        **kwargs,
    ):
        super().__init__(
            input_spec=kwargs.pop("input_spec"),
            output_spec=kwargs.pop("output_spec"),
            agent_group=kwargs.pop("agent_group"),
            input_has_agent_dim=kwargs.pop("input_has_agent_dim"),
            n_agents=kwargs.pop("n_agents"),
            centralised=kwargs.pop("centralised"),
            share_params=kwargs.pop("share_params"),
            device=kwargs.pop("device"),
            action_spec=kwargs.pop("action_spec"),
            model_index=kwargs.pop("model_index"),
            is_critic=kwargs.pop("is_critic"),
        )

        assert self.input_has_agent_dim is True

        out_dim = self.output_leaf_spec.shape[-1]
        mlp_out_dim = sum(spec[0] for spec in mlps_specs)
        k = out_dim // mlp_out_dim

        self.mlps_specs = mlps_specs
        input_sizes = [
            compute_input_size(self.input_spec.values(True, True), input_spec,
                               self.in_keys) for _, input_spec, _ in mlps_specs
        ]
        self.mlps = [
            MultiAgentMLP(n_agent_inputs=input_size,
                          n_agent_outputs=k * dims,
                          n_agents=self.n_agents,
                          centralised=self.centralised,
                          share_params=self.share_params,
                          device=self.device,
                          **ChainMap(i_kwargs, kwargs))
            for (dims, _, i_kwargs
                 ), input_size in zip(mlps_specs, input_sizes, strict=True)
        ]
        for i, mlp in enumerate(self.mlps):
            self.add_module(str(i), mlp)

    def _forward(self, tensordict: TensorDictBase) -> TensorDictBase:
        outs = []
        for mlp, (_, spec, _) in zip(self.mlps, self.mlps_specs, strict=True):
            input = compute_input(tensordict, spec, self.in_keys)
            out = mlp.forward(input)
            shape = out.shape
            outs.append(out)
        res = torch.stack(outs, dim=-1).reshape((*shape[:-1], -1))
        if not self.output_has_agent_dim:
            res = res[..., 0, :]
        tensordict.set(self.out_key, res)
        return tensordict


InputSpec = slice | Collection[str] | None


@dataclass
class SplitMlpConfig(ModelConfig):
    """
    A model with a independent MLP sub-models for different dimensions
    of the action space, as specified by ``mlps_specs``, a list of
    tuples ``(action_size, input_spec, net_arch)``
    each configure one of the MLPs that computes ``action_size`` actions
    using observations specified by``input_spec`` with an architecture ``net_arch``,
    similar to how :py:class:navground.learning.policies.SplitSACPolicy` is
    configured.

    The list should be ordered and actions sizes
    should sum up to the total size of the action space.


    For example:

    >>> env.observation_space('agent_0')
    Dict('a': Box(0.0, 1.0, (1,), float32), 'b': Box(0.0, 1.0, (1,), float32))
    >>> env.action_space('agent_0')
    Box(-1.0, 1.0, (2,), float32)
    >>> mlps_specs = [(1, None, {'num_cells': [64, 64]}), (1, ['a'],  {'num_cells': [16, 16]})]
    >>> model_config = SplitMlpConfig(mlps_specs=mlps_specs, ...)

    configures a model that computes two actions:

    - the first using a 64 + 64 MLP from ``a`` and ``b``
    - the second using a 16 + 16 MLP solely from ``a``

    """
    num_cells: Sequence[int] = MISSING  # type: ignore[assignment]
    layer_class: type[nn.Module] = MISSING  # type: ignore[assignment]

    activation_class: type[nn.Module] = MISSING  # type: ignore[assignment]
    activation_kwargs: dict | None = None

    norm_class: type[nn.Module] | None = None
    norm_kwargs: dict | None = None

    mlps_specs: list[tuple[int, InputSpec,
                           dict[str, Any]]] = field(default_factory=list)

    @staticmethod
    def associated_class():
        return SplitMlp


# https://pytorch.org/docs/main/optim.html#per-parameter-options


def select_submlp(experiment, index, lr=1e-4, eps=1e-6):
    split_mlp = experiment.policy[0][0][0]
    mlp = split_mlp.mlps[index]
    experiment.optimizers['agent']['loss_objective'] = torch.optim.Adam(
        mlp.parameters(), eps=eps, lr=lr)


def setup_optimizer(experiment, lrs, eps=1e-6):
    split_mlp = experiment.policy[0][0][0]
    params = [{
        'params': mlp.parameters(),
        'lr': lr
    } for mlp, lr in zip(split_mlp.mlps, lrs, strict=True)]
    experiment.optimizers['agent']['loss_objective'] = torch.optim.Adam(
        params, eps=eps)


def set_lr(experiment, index, lr):
    experiment.optimizers['agent']['loss_objective'].param_groups[index][
        'lr'] = lr


def compute_input_2(obs, i_spec, keys):
    if isinstance(obs, dict):
        if i_spec is None:
            return torch.cat([obs.get(in_key) for in_key in keys], dim=-1)
        return torch.cat(
            [obs.get(in_key) for in_key in keys if in_key in i_spec], dim=-1)
    if i_spec is None:
        return obs
    return obs[..., i_spec]


class SplitMlpPolicy(torch.nn.Module):

    def __init__(self, model: SplitMlp):
        super().__init__()
        self._specs = [x for _, x, _ in model.mlps_specs]
        self._mlp_nns = []
        self.in_keys = [x[-1] for x in model.in_keys]
        for mlp in model.mlps:
            mlp_nn = copy.deepcopy(mlp._empty_net)
            mlp.params.to_module(mlp_nn)
            self._mlp_nns.append(mlp_nn)
        for i, mlp in enumerate(self._mlp_nns):
            self.add_module(f'module_{i}', mlp)

    def forward(self, observation: PyTorchObs) -> torch.Tensor:
        logits = [
            mlp(compute_input_2(observation, spec, self.in_keys))
            for mlp, spec in zip(self._mlp_nns, self._specs, strict=True)
        ]
        return torch.stack(logits, dim=-1).reshape((*logits[0].shape[:-1], -1))



import math
from collections.abc import Callable
from functools import partial
from typing import Any

import numpy as np
from navground import core, sim
from navground.sim.ui import svg_color


def make_thymio() -> sim.Agent:
    kinematics = core.kinematics.TwoWheelsDifferentialDriveKinematics(
        wheel_axis=0.94,
        max_speed=0.166,
        max_forward_speed=0.14,
        max_backward_speed=0.14)
    agent = sim.Agent(radius=0.1, kinematics=kinematics)
    agent.type = "thymio"
    return agent


def are_two_agents_on_the_pad(world: sim.World,
                              tolerance: float | None = None) -> bool:
    if tolerance is None:
        tolerance = world.pad_tolerance  # type: ignore[attr-defined]
    x = world.pad_width / 2 - tolerance  # type: ignore[attr-defined]
    return len([a for a in world.agents if abs(a.position[0]) < x]) > 1


def draw_pad(world: sim.World) -> str:
    return '<rect fill="red" width="0.5" height="0.6" x="-0.25" y="-0.3" fill-opacity="0.2"/>'


def _draw_tx_world(world: sim.World,
                   low: float = -1,
                   high: float = 1,
                   color_low: tuple[float, float, float] = (0, 0, 0),
                   color_high: tuple[float, float, float] = (0, 0, 1),
                   stroke: str = 'white',
                   binarize: bool = False) -> str:
    r = ''
    for agent in world.agents:
        b = agent.behavior
        if b and hasattr(b, '_comm'):

            tx = float(b._comm)  # type: ignore[attr-defined]
            if binarize:
                if tx > (high + low) / 2:
                    color = svg_color(*color_high)
                else:
                    color = svg_color(*color_low)
            else:
                x = (tx - low) / (high - low)
                rgb = np.asarray(color_low) * (1 -
                                               x) + np.asarray(color_high) * x
                color = svg_color(*rgb)
            r += (
                f'<circle cx="{agent.position[0]}" cy="{agent.position[1]}" '
                f'r="0.025" fill="{color}" stroke="{stroke}" stroke-width="0.01"/>'
            )
    return r


def draw_tx(low: float = -1,
            high: float = 1,
            color_low: tuple[float, float, float] = (0, 0, 0),
            color_high: tuple[float, float, float] = (0, 0, 1),
            stroke: str = 'white',
            binarize: bool = False) -> Callable[[sim.World], str]:

    return partial(_draw_tx_world,
                   low=low,
                   high=high,
                   color_low=color_low,
                   color_high=color_high,
                   stroke=stroke,
                   binarize=binarize)


def color_on_pad(entity: sim.Entity, world: sim.World) -> dict[str, str]:
    if isinstance(entity, sim.Agent):
        if are_two_agents_on_the_pad(world):
            return {'fill': 'red'}
    return {}


class PadScenario(sim.Scenario, name="Pad"):

    def __init__(self,
                 length: float = 2,
                 width: float = 0.6,
                 with_walls: bool = True,
                 start_min_x: float = -math.inf,
                 start_max_x: float = math.inf,
                 start_in_opposite_sides: bool = True,
                 pad_tolerance: float = 0.01,
                 pad_width: float = 0.5):
        super().__init__()
        self._length = length
        self._width = width
        self._with_walls = with_walls
        self._start_min_x = start_min_x
        self._start_max_x = start_max_x
        self._start_in_opposite_sides = start_in_opposite_sides
        self._pad_tolerance = pad_tolerance
        self._pad_width = pad_width

    @property
    @sim.register(1.0, "Length")
    def length(self) -> float:
        return self._length

    @length.setter
    def length(self, value: float) -> None:
        self._length = max(0, value)

    @property
    @sim.register(0.6, "Width")
    def width(self) -> float:
        return self._width

    @width.setter
    def width(self, value: float) -> None:
        self._width = max(0, value)

    @property
    @sim.register(True, "Whether to add walls")
    def with_walls(self) -> bool:
        return self._with_walls

    @with_walls.setter
    def with_walls(self, value: bool) -> None:
        self._with_walls = value

    @property
    @sim.register(-math.inf, "Minimal starting position")
    def start_min_x(self) -> float:
        return self._start_min_x

    @start_min_x.setter
    def start_min_x(self, value: float) -> None:
        self._start_min_x = max(0, value)

    @property
    @sim.register(math.inf, "Maximal starting position")
    def start_max_x(self) -> float:
        return self._start_max_x

    @start_max_x.setter
    def start_max_x(self, value: float) -> None:
        self._start_max_x = max(0, value)

    @property
    @sim.register(True, "Whether to start in opposing sides")
    def start_in_opposite_sides(self) -> bool:
        return self._start_in_opposite_sides

    @start_in_opposite_sides.setter
    def start_in_opposite_sides(self, value: bool) -> None:
        self._start_in_opposite_sides = value

    @property
    @sim.register(0.01, "Tolerance")
    def pad_tolerance(self) -> float:
        return self._pad_tolerance

    @pad_tolerance.setter
    def pad_tolerance(self, value: float) -> None:
        self._pad_tolerance = max(0, value)

    @property
    @sim.register(0.5, "Pad width")
    def pad_width(self) -> float:
        return self._pad_width

    @pad_width.setter
    def pad_width(self, value: float) -> None:
        self._pad_width = max(0, value)

    def init_world(self, world: sim.World, seed: int | None = None) -> None:
        super().init_world(world, seed=seed)
        rng = world.random_generator
        world.bounding_box = sim.BoundingBox(-self.length / 2, self.length / 2,
                                             -self.width / 2, self.width / 2)
        if self.with_walls:
            for i in (-1, 1):
                world.add_wall(
                    sim.Wall((-10 * self.length, i * self.width / 2),
                             (10 * self.length, i * self.width / 2)))
        while len(world.agents) < 2:
            world.add_agent(self.default_agent())

        def terminate(world: sim.World):
            return all(
                agent.behavior.target.direction.dot(
                    agent.behavior.position) > self.length / 2
                for agent in world.agents if agent.behavior
                and agent.behavior.target.direction is not None)

        world.set_termination_condition(terminate)

        assert len(world.agents) == 2
        sides = -1, 1
        orientations = 0, np.pi
        colors = 'gold', 'cyan'
        min_x = max(self.start_min_x, -self.length / 2)
        max_x = min(self.start_max_x, self.length / 2)
        if self.start_in_opposite_sides:
            min_x = max(min_x, self.pad_width / 2 + self.pad_tolerance)
        y = self.width / 4
        for agent, side, orientation, color in zip(world.agents,
                                                   sides,
                                                   orientations,
                                                   colors,
                                                   strict=True):
            if not agent.task:
                agent.task = sim.tasks.DirectionTask(direction=(-side, 0))
            if not agent.behavior:
                agent.behavior = core.behaviors.DummyBehavior()
            if not agent.kinematics:
                agent.kinematics = core.kinematics.TwoWheelsDifferentialDriveKinematics(
                    wheel_axis=0.94,
                    max_speed=0.166,
                    max_forward_speed=0.14,
                    max_backward_speed=0.14)
            if not agent.type:
                agent.type = "thymio"
            if not agent.radius:
                agent.radius = 0.1
            if not agent.color:
                agent.color = color
            agent.pose = core.Pose2(
                (side * rng.uniform(min_x, max_x), side * y), orientation)
        world.pad_width = self.pad_width  # type: ignore[attr-defined]
        world.pad_tolerance = self.pad_tolerance  # type: ignore[attr-defined]
        world.set("pad_width", self.pad_width)
        world.set("pad_tolerance", self.pad_tolerance)

    def default_agent(self) -> sim.Agent:
        return make_thymio()


def render_kwargs(comm: bool = False,
                  low: float = -1,
                  high: float = 1,
                  color_low: tuple[float, float, float] = (0, 0, 0),
                  color_high: tuple[float, float, float] = (0, 0, 1),
                  stroke: str = 'white',
                  binarize: bool = False) -> dict[str, Any]:
    """
    The kwargs to pass to the world renderer

    :param      comm: Whether to display transmitted messages using a LED.
    :param      low:  The message lower bound
    :param      high:  The message upper bound
    :param      color_low: The color associated to lower-valued messages
    :param      color_high: The color associated to higher-valued messages
    :param      stroke: The color of the LED boundary
    :param      binarize: Whether to binary message values.

    :returns:   The kwargs
    """
    rs = {'background_extras': [draw_pad], 'decorate': color_on_pad}
    if comm:
        rs['extras'] = [
            draw_tx(low=low,
                    high=high,
                    color_low=color_low,
                    color_high=color_high,
                    stroke=stroke,
                    binarize=binarize)
        ]

    return rs



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from .env import BaseParallelEnv
import types

if TYPE_CHECKING:
    from stable_baselines3.common.vec_env import VecEnv

def make_vec_from_penv(
    env: BaseParallelEnv,
    num_envs: int = 1,
    processes: int = 1,
    black_death: bool = False,
    seed: int = 0,
    monitor: bool = False,
    monitor_keywords: tuple[str] = ("is_success", )
) -> VecEnv:
    """
    Converts a :py:class:`PettingZoo  Parallel environment
    <pettingzoo.utils.env.ParallelEnv>`
    to a :py:class:`StableBaseline3 vectorized environment
    <stable_baselines3.common.vec_env.VecEnv>`

    The multiple agents of the single PettingZoo
    (action/observation spaces needs to be shared)
    are stuck as multiple environments of a single agent using
    :py:func:`supersuit.pettingzoo_env_to_vec_env_v1`, and then concatenated
    using :py:func:`supersuit.concat_vec_envs_v1`.

    :param      env:        The environment
    :param      num_envs:   The number of parallel envs to concatenate
    :param      processes:  The number of processes
    :param      seed:       The seed
    :param      black_death: Whether to allow dynamic number of agents
    :param      monitor: Whether to wrap the vector env in a ``VecMonitor``
    :param      monitor_keywords: The keywords passed to ``VecMonitor``

    :returns:   The vector environment with ``number x |agents|``
                single agent environments.
    """
    import supersuit  # type: ignore[import-untyped]
    from stable_baselines3.common.vec_env import VecMonitor

    # penv = supersuit.pettingzoo_env_to_vec_env_v1(env)
    menv = supersuit.vector.MarkovVectorEnv(env, black_death=black_death)
    venv = supersuit.concat_vec_envs_v1(menv,
                                        num_envs,
                                        num_cpus=processes,
                                        base_class="stable_baselines3")

    def get_attr(self, attr_name: str, indices=None) -> list:
        if indices is None:
            indices = range(num_envs * menv.num_envs)
        if isinstance(indices, int):
            indices = [indices]
        rs = []
        for i in indices:
            penv = self.vec_envs[i // menv.num_envs].par_env
            rs.append(getattr(penv, attr_name))
        if attr_name == 'render_mode':
            # This is an attribute of vector environments themselves
            return rs[0]
        return rs

    def set_attr(self, attr_name: str, value, indices=None) -> None:
        if indices is None:
            indices = range(num_envs * menv.num_envs)
        if isinstance(indices, int):
            indices = [indices]
        for i in indices:
            penv = self.vec_envs[i // menv.num_envs].par_env
            setattr(penv, attr_name, value)

    cenv = venv.unwrapped
    cenv.get_attr = types.MethodType(get_attr, cenv)
    cenv.set_attr = types.MethodType(set_attr, cenv)

    for i, env in enumerate(venv.venv.vec_envs):
        env.reset(seed + i)
    if monitor:
        venv = VecMonitor(venv, info_keywords=monitor_keywords)
    return cast("VecEnv", venv)
